# Copyright (c) 2022-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

import pytest

from swh.graph.example_dataset import DATASET_DIR
from swh.graph.grpc.swhgraph_pb2 import GetNodeRequest, GraphDirection

# locally "redefine" all fixtures that depend on the session-scoped
# graph_grpc_server_config, because we need pytest to call them again.
from swh.graph.pytest_plugin import (  # noqa
    graph_grpc_server_process,
    graph_grpc_server_started,
)

from .test_getnode import TEST_ORIGIN_ID, test_invalid_swhid, test_not_found  # noqa
from .test_neighbors import traverse_nodes


@pytest.fixture(scope="session")
def graph_grpc_server_config(
    graph_grpc_backend_implementation, graph_statsd_server, tmpdir_factory
):
    full_graph_dir = DATASET_DIR / "compressed"
    partial_graph_path = Path(tmpdir_factory.mktemp("partial-graph"))
    for filepath in full_graph_dir.iterdir():
        # copy everything but property files
        if filepath.name.startswith("example.property."):
            continue
        symlink_path = partial_graph_path / filepath.name
        symlink_path.symlink_to(filepath)

    return {
        "graph": {
            "cls": f"local_{graph_grpc_backend_implementation}",
            "grpc_server": {
                "path": partial_graph_path / "example",
                "debug": True,
                "statsd_host": graph_statsd_server.host,
                "statsd_port": graph_statsd_server.port,
            },
            "http_rpc_server": {"debug": True},
        }
    }


def test_cnt(graph_grpc_stub):
    swhid = "swh:1:cnt:0000000000000000000000000000000000000001"

    node = graph_grpc_stub.GetNode(GetNodeRequest(swhid=swhid))
    if node.HasField("cnt"):
        assert not node.cnt.HasField("length")
        assert not node.cnt.HasField("is_skipped")


def test_rev(graph_grpc_stub):
    swhid = "swh:1:rev:0000000000000000000000000000000000000003"

    node = graph_grpc_stub.GetNode(GetNodeRequest(swhid=swhid))
    assert not node.rev.HasField("message")
    assert not node.rev.HasField("author")
    assert not node.rev.HasField("author_date")
    assert not node.rev.HasField("author_date_offset")
    assert not node.rev.HasField("committer")
    assert not node.rev.HasField("committer_date")
    assert not node.rev.HasField("committer_date_offset")


@pytest.mark.parametrize("rel_id", [10, 19])
def test_rel(rel_id, graph_grpc_stub):
    swhid = f"swh:1:rel:00000000000000000000000000000000000000{rel_id}"

    node = graph_grpc_stub.GetNode(GetNodeRequest(swhid=swhid))
    assert not node.rel.HasField("message")
    assert not node.rel.HasField("author")
    assert not node.rel.HasField("author_date")
    assert not node.rel.HasField("author_date_offset")


def test_ori(graph_grpc_stub):
    swhid = TEST_ORIGIN_ID

    node = graph_grpc_stub.GetNode(GetNodeRequest(swhid=swhid))
    assert not node.ori.HasField("url")


def test_neighbors(graph_grpc_stub):
    expected_nodes = []

    actual_nodes = traverse_nodes(
        graph_grpc_stub, ["swh:1:rev:0000000000000000000000000000000000000003"]
    )
    expected_nodes = ["swh:1:dir:0000000000000000000000000000000000000002"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:dir:0000000000000000000000000000000000000017"],
        edges="dir:cnt",
    )
    expected_nodes = ["swh:1:cnt:0000000000000000000000000000000000000014"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:dir:0000000000000000000000000000000000000012"],
        direction=GraphDirection.BACKWARD,
    )
    expected_nodes = ["swh:1:rev:0000000000000000000000000000000000000013"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:rev:0000000000000000000000000000000000000009"],
        direction=GraphDirection.BACKWARD,
        edges="rev:rev",
    )
    expected_nodes = ["swh:1:rev:0000000000000000000000000000000000000013"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:snp:0000000000000000000000000000000000000020"],
        direction=GraphDirection.BACKWARD,
    )
    expected_nodes = [TEST_ORIGIN_ID]
    assert set(actual_nodes) == set(expected_nodes)



#!/usr/bin/env python3

# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# type: ignore

import argparse
import logging
from pathlib import Path
import shutil

from swh.export.exporters.edges import GraphEdgesExporter
from swh.export.exporters.orc import ORCExporter
from swh.graph.example_dataset import DATASET
from swh.graph.webgraph import compress


def main():
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(description="Generate a test dataset")
    parser.add_argument(
        "--compress",
        action="store_true",
        default=False,
        help="Also compress the dataset",
    )
    parser.add_argument(
        "--profile", default="release", help="rust profile to use for compression"
    )
    parser.add_argument("output", help="output directory", nargs="?", default=".")
    args = parser.parse_args()

    exporters = {"edges": GraphEdgesExporter, "orc": ORCExporter}
    config = {"test_unique_file_id": "all"}
    output_path = Path(args.output)
    for name, exporter in exporters.items():
        if (output_path / name).exists():
            shutil.rmtree(output_path / name)
        with exporter(config, output_path / name) as e:
            for obj in DATASET:
                e.process_object(obj.object_type, obj)

    if args.compress:
        if (output_path / "compressed").exists():
            shutil.rmtree(output_path / "compressed")
        compress(
            "example",
            output_path / "orc",
            output_path / "compressed",
            "example",
            conf={"profile": args.profile},
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2022-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

from google.protobuf.field_mask_pb2 import FieldMask
import pytest

from swh.graph.example_dataset import DATASET_DIR
from swh.graph.grpc.swhgraph_pb2 import (
    GraphDirection,
    Node,
    Successor,
    TraversalRequest,
)

# locally "redefine" all fixtures that depend on the session-scoped
# graph_grpc_server_config, because we need pytest to call them again.
from swh.graph.pytest_plugin import (  # noqa
    graph_grpc_server_process,
    graph_grpc_server_started,
)

from .test_getnode import (  # noqa
    TEST_ORIGIN_ID,
    TEST_ORIGIN_ID2,
    test_invalid_swhid,
    test_not_found,
)
from .test_neighbors import traverse_nodes


@pytest.fixture(scope="session")
def graph_grpc_server_config(
    graph_grpc_backend_implementation, graph_statsd_server, tmpdir_factory
):
    full_graph_dir = DATASET_DIR / "compressed"
    partial_graph_path = Path(tmpdir_factory.mktemp("partial-graph"))
    for filepath in full_graph_dir.iterdir():
        # copy everything but labels
        if "-labelled" in filepath.name:
            continue
        symlink_path = partial_graph_path / filepath.name
        symlink_path.symlink_to(filepath)

    return {
        "graph": {
            "cls": f"local_{graph_grpc_backend_implementation}",
            "grpc_server": {
                "path": partial_graph_path / "example",
                "extra_options": ["--labels", "none"],
                "debug": True,
                "statsd_host": graph_statsd_server.host,
                "statsd_port": graph_statsd_server.port,
            },
            "http_rpc_server": {"debug": True},
        }
    }


def test_neighbors(graph_grpc_stub):
    expected_nodes = []

    actual_nodes = traverse_nodes(
        graph_grpc_stub, ["swh:1:rev:0000000000000000000000000000000000000003"]
    )
    expected_nodes = ["swh:1:dir:0000000000000000000000000000000000000002"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:dir:0000000000000000000000000000000000000017"],
        edges="dir:cnt",
    )
    expected_nodes = ["swh:1:cnt:0000000000000000000000000000000000000014"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:dir:0000000000000000000000000000000000000012"],
        direction=GraphDirection.BACKWARD,
    )
    expected_nodes = ["swh:1:rev:0000000000000000000000000000000000000013"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:rev:0000000000000000000000000000000000000009"],
        direction=GraphDirection.BACKWARD,
        edges="rev:rev",
    )
    expected_nodes = ["swh:1:rev:0000000000000000000000000000000000000013"]
    assert set(actual_nodes) == set(expected_nodes)

    actual_nodes = traverse_nodes(
        graph_grpc_stub,
        ["swh:1:snp:0000000000000000000000000000000000000020"],
        direction=GraphDirection.BACKWARD,
    )
    expected_nodes = [TEST_ORIGIN_ID]
    assert set(actual_nodes) == set(expected_nodes)


def test_traverse_forward_labels(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=[TEST_ORIGIN_ID],
            mask=FieldMask(paths=["swhid", "successor.swhid", "successor.label"]),
            direction=GraphDirection.FORWARD,
        )
    )
    expected = [
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000001",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000004",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000005",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000007",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000002",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=None,
                )
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000006",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000004",
                    label=None,
                ),
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000005",
                    label=None,
                ),
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000008",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=None,
                ),
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000007",
                    label=None,
                ),
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000006",
                    label=None,
                ),
            ],
        ),
        Node(
            swhid=TEST_ORIGIN_ID,
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000020",
                    label=None,
                ),
            ],
            ori=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000010",
            successor=[
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000009",
                    label=[],
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000003",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000002",
                    label=[],
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000009",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000008",
                    label=[],
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000003",
                    label=[],
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000020",
            successor=[
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000010",
                    label=None,
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000009",
                    label=None,
                ),
            ],
        ),
    ]
    actual = list(request)
    actual.sort(key=lambda node: node.swhid)
    for node in actual:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert actual == expected


def test_traverse_backward_labels(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:cnt:0000000000000000000000000000000000000015"],
            mask=FieldMask(paths=["swhid", "successor.swhid", "successor.label"]),
            direction=GraphDirection.BACKWARD,
        )
    )
    expected = [
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000015",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000016",
                    label=None,
                )
            ],
            cnt=None,
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000016",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000017",
                    label=None,
                ),
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000017",
            successor=[
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000018",
                    label=[],
                ),
            ],
        ),
        Node(
            swhid=TEST_ORIGIN_ID2,
            successor=None,
            ori=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000019",
            successor=[],
            rel=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000021",
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000022",
                    label=None,
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000018",
            successor=[
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000019",
                    label=None,
                ),
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000021",
                    label=None,
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000022",
            successor=[
                Successor(
                    swhid=TEST_ORIGIN_ID2,
                    label=None,
                ),
            ],
        ),
    ]
    actual = list(request)
    actual.sort(key=lambda node: node.swhid)
    for node in actual:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert actual == expected



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.field_mask_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _GraphDirection:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _GraphDirectionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_GraphDirection.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    FORWARD: _GraphDirection.ValueType  # 0
    """Forward DAG: ori -> snp -> rel -> rev -> dir -> cnt"""

    BACKWARD: _GraphDirection.ValueType  # 1
    """Transposed DAG: cnt -> dir -> rev -> rel -> snp -> ori"""

class GraphDirection(_GraphDirection, metaclass=_GraphDirectionEnumTypeWrapper):
    """Direction of the graph"""
    pass

FORWARD: GraphDirection.ValueType  # 0
"""Forward DAG: ori -> snp -> rel -> rev -> dir -> cnt"""

BACKWARD: GraphDirection.ValueType  # 1
"""Transposed DAG: cnt -> dir -> rev -> rel -> snp -> ori"""

global___GraphDirection = GraphDirection


class GetNodeRequest(google.protobuf.message.Message):
    """Describe a node to return"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SWHID_FIELD_NUMBER: builtins.int
    MASK_FIELD_NUMBER: builtins.int
    swhid: typing.Text
    """SWHID of the node to return"""

    @property
    def mask(self) -> google.protobuf.field_mask_pb2.FieldMask:
        """FieldMask of which fields are to be returned (e.g., "swhid,cnt.length").
        By default, all fields are returned.
        """
        pass
    def __init__(self,
        *,
        swhid: typing.Text = ...,
        mask: typing.Optional[google.protobuf.field_mask_pb2.FieldMask] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_mask",b"_mask","mask",b"mask"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_mask",b"_mask","mask",b"mask","swhid",b"swhid"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_mask",b"_mask"]) -> typing.Optional[typing_extensions.Literal["mask"]]: ...
global___GetNodeRequest = GetNodeRequest

class TraversalRequest(google.protobuf.message.Message):
    """TraversalRequest describes how a breadth-first traversal should be
    performed, and what should be returned to the client.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SRC_FIELD_NUMBER: builtins.int
    DIRECTION_FIELD_NUMBER: builtins.int
    EDGES_FIELD_NUMBER: builtins.int
    MAX_EDGES_FIELD_NUMBER: builtins.int
    MIN_DEPTH_FIELD_NUMBER: builtins.int
    MAX_DEPTH_FIELD_NUMBER: builtins.int
    RETURN_NODES_FIELD_NUMBER: builtins.int
    MASK_FIELD_NUMBER: builtins.int
    MAX_MATCHING_NODES_FIELD_NUMBER: builtins.int
    @property
    def src(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Set of source nodes (SWHIDs)"""
        pass
    direction: global___GraphDirection.ValueType
    """Direction of the graph to traverse. Defaults to FORWARD."""

    edges: typing.Text
    """Edge restriction string (e.g. "rev:dir,dir:cnt").
    Defaults to "*" (all).
    """

    max_edges: builtins.int
    """Maximum number of edges accessed in the traversal, after which it stops.
    Defaults to infinite.
    """

    min_depth: builtins.int
    """Do not return nodes with a depth lower than this number.
    By default, all depths are returned.
    """

    max_depth: builtins.int
    """Maximum depth of the traversal, after which it stops.
    Defaults to infinite.
    """

    @property
    def return_nodes(self) -> global___NodeFilter:
        """Filter which nodes will be sent to the stream. By default, all nodes are
        returned.
        """
        pass
    @property
    def mask(self) -> google.protobuf.field_mask_pb2.FieldMask:
        """FieldMask of which fields are to be returned (e.g., "swhid,cnt.length").
        By default, all fields are returned.
        """
        pass
    max_matching_nodes: builtins.int
    """Maximum number of matching results before stopping. For Traverse(), this is
    the total number of results. Defaults to infinite.
    """

    def __init__(self,
        *,
        src: typing.Optional[typing.Iterable[typing.Text]] = ...,
        direction: global___GraphDirection.ValueType = ...,
        edges: typing.Optional[typing.Text] = ...,
        max_edges: typing.Optional[builtins.int] = ...,
        min_depth: typing.Optional[builtins.int] = ...,
        max_depth: typing.Optional[builtins.int] = ...,
        return_nodes: typing.Optional[global___NodeFilter] = ...,
        mask: typing.Optional[google.protobuf.field_mask_pb2.FieldMask] = ...,
        max_matching_nodes: typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_edges",b"_edges","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","_max_matching_nodes",b"_max_matching_nodes","_min_depth",b"_min_depth","_return_nodes",b"_return_nodes","edges",b"edges","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges","max_matching_nodes",b"max_matching_nodes","min_depth",b"min_depth","return_nodes",b"return_nodes"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_edges",b"_edges","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","_max_matching_nodes",b"_max_matching_nodes","_min_depth",b"_min_depth","_return_nodes",b"_return_nodes","direction",b"direction","edges",b"edges","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges","max_matching_nodes",b"max_matching_nodes","min_depth",b"min_depth","return_nodes",b"return_nodes","src",b"src"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_edges",b"_edges"]) -> typing.Optional[typing_extensions.Literal["edges"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_mask",b"_mask"]) -> typing.Optional[typing_extensions.Literal["mask"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_depth",b"_max_depth"]) -> typing.Optional[typing_extensions.Literal["max_depth"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_edges",b"_max_edges"]) -> typing.Optional[typing_extensions.Literal["max_edges"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_matching_nodes",b"_max_matching_nodes"]) -> typing.Optional[typing_extensions.Literal["max_matching_nodes"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_min_depth",b"_min_depth"]) -> typing.Optional[typing_extensions.Literal["min_depth"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_return_nodes",b"_return_nodes"]) -> typing.Optional[typing_extensions.Literal["return_nodes"]]: ...
global___TraversalRequest = TraversalRequest

class FindPathToRequest(google.protobuf.message.Message):
    """FindPathToRequest describes a request to find a shortest path between a
    set of nodes and a given target criteria, as well as what should be returned
    in the path.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SRC_FIELD_NUMBER: builtins.int
    TARGET_FIELD_NUMBER: builtins.int
    DIRECTION_FIELD_NUMBER: builtins.int
    EDGES_FIELD_NUMBER: builtins.int
    MAX_EDGES_FIELD_NUMBER: builtins.int
    MAX_DEPTH_FIELD_NUMBER: builtins.int
    MASK_FIELD_NUMBER: builtins.int
    @property
    def src(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Set of source nodes (SWHIDs)"""
        pass
    @property
    def target(self) -> global___NodeFilter:
        """Target criteria, i.e., what constitutes a valid path destination."""
        pass
    direction: global___GraphDirection.ValueType
    """Direction of the graph to traverse. Defaults to FORWARD."""

    edges: typing.Text
    """Edge restriction string (e.g. "rev:dir,dir:cnt").
    Defaults to "*" (all).
    """

    max_edges: builtins.int
    """Maximum number of edges accessed in the traversal, after which it stops.
    Defaults to infinite.
    """

    max_depth: builtins.int
    """Maximum depth of the traversal, after which it stops.
    Defaults to infinite.
    """

    @property
    def mask(self) -> google.protobuf.field_mask_pb2.FieldMask:
        """FieldMask of which fields are to be returned (e.g., "swhid,cnt.length").
        By default, all fields are returned.
        """
        pass
    def __init__(self,
        *,
        src: typing.Optional[typing.Iterable[typing.Text]] = ...,
        target: typing.Optional[global___NodeFilter] = ...,
        direction: global___GraphDirection.ValueType = ...,
        edges: typing.Optional[typing.Text] = ...,
        max_edges: typing.Optional[builtins.int] = ...,
        max_depth: typing.Optional[builtins.int] = ...,
        mask: typing.Optional[google.protobuf.field_mask_pb2.FieldMask] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_edges",b"_edges","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","edges",b"edges","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges","target",b"target"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_edges",b"_edges","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","direction",b"direction","edges",b"edges","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges","src",b"src","target",b"target"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_edges",b"_edges"]) -> typing.Optional[typing_extensions.Literal["edges"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_mask",b"_mask"]) -> typing.Optional[typing_extensions.Literal["mask"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_depth",b"_max_depth"]) -> typing.Optional[typing_extensions.Literal["max_depth"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_edges",b"_max_edges"]) -> typing.Optional[typing_extensions.Literal["max_edges"]]: ...
global___FindPathToRequest = FindPathToRequest

class FindPathBetweenRequest(google.protobuf.message.Message):
    """FindPathToRequest describes a request to find a shortest path between a
    set of source nodes and a set of destination nodes. It works by performing a
    bidirectional breadth-first traversal from both sets at the same time.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SRC_FIELD_NUMBER: builtins.int
    DST_FIELD_NUMBER: builtins.int
    DIRECTION_FIELD_NUMBER: builtins.int
    DIRECTION_REVERSE_FIELD_NUMBER: builtins.int
    EDGES_FIELD_NUMBER: builtins.int
    EDGES_REVERSE_FIELD_NUMBER: builtins.int
    MAX_EDGES_FIELD_NUMBER: builtins.int
    MAX_DEPTH_FIELD_NUMBER: builtins.int
    MASK_FIELD_NUMBER: builtins.int
    @property
    def src(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Set of source nodes (SWHIDs)"""
        pass
    @property
    def dst(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Set of destination nodes (SWHIDs)"""
        pass
    direction: global___GraphDirection.ValueType
    """Direction of the graph to traverse from the source set. Defaults to
    FORWARD.
    """

    direction_reverse: global___GraphDirection.ValueType
    """Direction of the graph to traverse from the destination set. Defaults to
    the opposite of `direction`. If direction and direction_reverse are
    identical, it will find the first common successor of both sets in the
    given direction.
    """

    edges: typing.Text
    """Edge restriction string for the traversal from the source set.
    (e.g. "rev:dir,dir:cnt"). Defaults to "*" (all).
    """

    edges_reverse: typing.Text
    """Edge restriction string for the reverse traversal from the destination
    set.
    If not specified:
      - If `edges` is not specified either, defaults to "*"
      - If direction == direction_reverse, defaults to `edges`
      - If direction != direction_reverse, defaults to the reverse of `edges`
        (e.g. "rev:dir" becomes "dir:rev").
    """

    max_edges: builtins.int
    """Maximum number of edges accessed in the traversal, after which it stops.
    Defaults to infinite.
    """

    max_depth: builtins.int
    """Maximum depth of the traversal, after which it stops.
    Defaults to infinite.
    """

    @property
    def mask(self) -> google.protobuf.field_mask_pb2.FieldMask:
        """FieldMask of which fields are to be returned (e.g., "swhid,cnt.length").
        By default, all fields are returned.
        """
        pass
    def __init__(self,
        *,
        src: typing.Optional[typing.Iterable[typing.Text]] = ...,
        dst: typing.Optional[typing.Iterable[typing.Text]] = ...,
        direction: global___GraphDirection.ValueType = ...,
        direction_reverse: typing.Optional[global___GraphDirection.ValueType] = ...,
        edges: typing.Optional[typing.Text] = ...,
        edges_reverse: typing.Optional[typing.Text] = ...,
        max_edges: typing.Optional[builtins.int] = ...,
        max_depth: typing.Optional[builtins.int] = ...,
        mask: typing.Optional[google.protobuf.field_mask_pb2.FieldMask] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_direction_reverse",b"_direction_reverse","_edges",b"_edges","_edges_reverse",b"_edges_reverse","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","direction_reverse",b"direction_reverse","edges",b"edges","edges_reverse",b"edges_reverse","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_direction_reverse",b"_direction_reverse","_edges",b"_edges","_edges_reverse",b"_edges_reverse","_mask",b"_mask","_max_depth",b"_max_depth","_max_edges",b"_max_edges","direction",b"direction","direction_reverse",b"direction_reverse","dst",b"dst","edges",b"edges","edges_reverse",b"edges_reverse","mask",b"mask","max_depth",b"max_depth","max_edges",b"max_edges","src",b"src"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_direction_reverse",b"_direction_reverse"]) -> typing.Optional[typing_extensions.Literal["direction_reverse"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_edges",b"_edges"]) -> typing.Optional[typing_extensions.Literal["edges"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_edges_reverse",b"_edges_reverse"]) -> typing.Optional[typing_extensions.Literal["edges_reverse"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_mask",b"_mask"]) -> typing.Optional[typing_extensions.Literal["mask"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_depth",b"_max_depth"]) -> typing.Optional[typing_extensions.Literal["max_depth"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_edges",b"_max_edges"]) -> typing.Optional[typing_extensions.Literal["max_edges"]]: ...
global___FindPathBetweenRequest = FindPathBetweenRequest

class NodeFilter(google.protobuf.message.Message):
    """Represents various criteria that make a given node "valid". A node is
    only valid if all the subcriteria present in this message are fulfilled.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TYPES_FIELD_NUMBER: builtins.int
    MIN_TRAVERSAL_SUCCESSORS_FIELD_NUMBER: builtins.int
    MAX_TRAVERSAL_SUCCESSORS_FIELD_NUMBER: builtins.int
    types: typing.Text
    """Node restriction string. (e.g. "dir,cnt,rev"). Defaults to "*" (all)."""

    min_traversal_successors: builtins.int
    """Minimum number of successors encountered *during the traversal*.
    Default: no constraint
    """

    max_traversal_successors: builtins.int
    """Maximum number of successors encountered *during the traversal*.
    Default: no constraint
    """

    def __init__(self,
        *,
        types: typing.Optional[typing.Text] = ...,
        min_traversal_successors: typing.Optional[builtins.int] = ...,
        max_traversal_successors: typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_max_traversal_successors",b"_max_traversal_successors","_min_traversal_successors",b"_min_traversal_successors","_types",b"_types","max_traversal_successors",b"max_traversal_successors","min_traversal_successors",b"min_traversal_successors","types",b"types"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_max_traversal_successors",b"_max_traversal_successors","_min_traversal_successors",b"_min_traversal_successors","_types",b"_types","max_traversal_successors",b"max_traversal_successors","min_traversal_successors",b"min_traversal_successors","types",b"types"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_max_traversal_successors",b"_max_traversal_successors"]) -> typing.Optional[typing_extensions.Literal["max_traversal_successors"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_min_traversal_successors",b"_min_traversal_successors"]) -> typing.Optional[typing_extensions.Literal["min_traversal_successors"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_types",b"_types"]) -> typing.Optional[typing_extensions.Literal["types"]]: ...
global___NodeFilter = NodeFilter

class Node(google.protobuf.message.Message):
    """Represents a node in the graph."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SWHID_FIELD_NUMBER: builtins.int
    SUCCESSOR_FIELD_NUMBER: builtins.int
    NUM_SUCCESSORS_FIELD_NUMBER: builtins.int
    CNT_FIELD_NUMBER: builtins.int
    REV_FIELD_NUMBER: builtins.int
    REL_FIELD_NUMBER: builtins.int
    ORI_FIELD_NUMBER: builtins.int
    swhid: typing.Text
    """The SWHID of the graph node."""

    @property
    def successor(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Successor]:
        """List of relevant successors of this node."""
        pass
    num_successors: builtins.int
    """Number of relevant successors."""

    @property
    def cnt(self) -> global___ContentData: ...
    @property
    def rev(self) -> global___RevisionData: ...
    @property
    def rel(self) -> global___ReleaseData: ...
    @property
    def ori(self) -> global___OriginData: ...
    def __init__(self,
        *,
        swhid: typing.Text = ...,
        successor: typing.Optional[typing.Iterable[global___Successor]] = ...,
        num_successors: typing.Optional[builtins.int] = ...,
        cnt: typing.Optional[global___ContentData] = ...,
        rev: typing.Optional[global___RevisionData] = ...,
        rel: typing.Optional[global___ReleaseData] = ...,
        ori: typing.Optional[global___OriginData] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_num_successors",b"_num_successors","cnt",b"cnt","data",b"data","num_successors",b"num_successors","ori",b"ori","rel",b"rel","rev",b"rev"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_num_successors",b"_num_successors","cnt",b"cnt","data",b"data","num_successors",b"num_successors","ori",b"ori","rel",b"rel","rev",b"rev","successor",b"successor","swhid",b"swhid"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_num_successors",b"_num_successors"]) -> typing.Optional[typing_extensions.Literal["num_successors"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data",b"data"]) -> typing.Optional[typing_extensions.Literal["cnt","rev","rel","ori"]]: ...
global___Node = Node

class Path(google.protobuf.message.Message):
    """Represents a path in the graph."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NODE_FIELD_NUMBER: builtins.int
    MIDPOINT_INDEX_FIELD_NUMBER: builtins.int
    @property
    def node(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Node]:
        """List of nodes in the path, from source to destination"""
        pass
    midpoint_index: builtins.int
    """Index of the "midpoint" of the path. For paths obtained with
    bidirectional search queries, this is the node that joined the two
    sets together. When looking for a common ancestor between two nodes by
    performing a FindPathBetween search with two backward graphs, this will
    be the index of the common ancestor in the path.
    """

    def __init__(self,
        *,
        node: typing.Optional[typing.Iterable[global___Node]] = ...,
        midpoint_index: typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_midpoint_index",b"_midpoint_index","midpoint_index",b"midpoint_index"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_midpoint_index",b"_midpoint_index","midpoint_index",b"midpoint_index","node",b"node"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_midpoint_index",b"_midpoint_index"]) -> typing.Optional[typing_extensions.Literal["midpoint_index"]]: ...
global___Path = Path

class Successor(google.protobuf.message.Message):
    """Represents a successor of a given node."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SWHID_FIELD_NUMBER: builtins.int
    LABEL_FIELD_NUMBER: builtins.int
    swhid: typing.Text
    """The SWHID of the successor"""

    @property
    def label(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___EdgeLabel]:
        """A list of edge labels for the given edge"""
        pass
    def __init__(self,
        *,
        swhid: typing.Optional[typing.Text] = ...,
        label: typing.Optional[typing.Iterable[global___EdgeLabel]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_swhid",b"_swhid","swhid",b"swhid"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_swhid",b"_swhid","label",b"label","swhid",b"swhid"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_swhid",b"_swhid"]) -> typing.Optional[typing_extensions.Literal["swhid"]]: ...
global___Successor = Successor

class ContentData(google.protobuf.message.Message):
    """Content node properties"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    LENGTH_FIELD_NUMBER: builtins.int
    IS_SKIPPED_FIELD_NUMBER: builtins.int
    length: builtins.int
    """Length of the blob, in bytes"""

    is_skipped: builtins.bool
    """Whether the content was skipped during ingestion."""

    def __init__(self,
        *,
        length: typing.Optional[builtins.int] = ...,
        is_skipped: typing.Optional[builtins.bool] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_is_skipped",b"_is_skipped","_length",b"_length","is_skipped",b"is_skipped","length",b"length"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_is_skipped",b"_is_skipped","_length",b"_length","is_skipped",b"is_skipped","length",b"length"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_is_skipped",b"_is_skipped"]) -> typing.Optional[typing_extensions.Literal["is_skipped"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_length",b"_length"]) -> typing.Optional[typing_extensions.Literal["length"]]: ...
global___ContentData = ContentData

class RevisionData(google.protobuf.message.Message):
    """Revision node properties"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTHOR_FIELD_NUMBER: builtins.int
    AUTHOR_DATE_FIELD_NUMBER: builtins.int
    AUTHOR_DATE_OFFSET_FIELD_NUMBER: builtins.int
    COMMITTER_FIELD_NUMBER: builtins.int
    COMMITTER_DATE_FIELD_NUMBER: builtins.int
    COMMITTER_DATE_OFFSET_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    author: builtins.int
    """Revision author ID (anonymized)"""

    author_date: builtins.int
    """UNIX timestamp of the revision date (UTC)"""

    author_date_offset: builtins.int
    """Timezone of the revision author date as an offset from UTC"""

    committer: builtins.int
    """Revision committer ID (anonymized)"""

    committer_date: builtins.int
    """UNIX timestamp of the revision committer date (UTC)"""

    committer_date_offset: builtins.int
    """Timezone of the revision committer date as an offset from UTC"""

    message: builtins.bytes
    """Revision message"""

    def __init__(self,
        *,
        author: typing.Optional[builtins.int] = ...,
        author_date: typing.Optional[builtins.int] = ...,
        author_date_offset: typing.Optional[builtins.int] = ...,
        committer: typing.Optional[builtins.int] = ...,
        committer_date: typing.Optional[builtins.int] = ...,
        committer_date_offset: typing.Optional[builtins.int] = ...,
        message: typing.Optional[builtins.bytes] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_author",b"_author","_author_date",b"_author_date","_author_date_offset",b"_author_date_offset","_committer",b"_committer","_committer_date",b"_committer_date","_committer_date_offset",b"_committer_date_offset","_message",b"_message","author",b"author","author_date",b"author_date","author_date_offset",b"author_date_offset","committer",b"committer","committer_date",b"committer_date","committer_date_offset",b"committer_date_offset","message",b"message"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_author",b"_author","_author_date",b"_author_date","_author_date_offset",b"_author_date_offset","_committer",b"_committer","_committer_date",b"_committer_date","_committer_date_offset",b"_committer_date_offset","_message",b"_message","author",b"author","author_date",b"author_date","author_date_offset",b"author_date_offset","committer",b"committer","committer_date",b"committer_date","committer_date_offset",b"committer_date_offset","message",b"message"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author",b"_author"]) -> typing.Optional[typing_extensions.Literal["author"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author_date",b"_author_date"]) -> typing.Optional[typing_extensions.Literal["author_date"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author_date_offset",b"_author_date_offset"]) -> typing.Optional[typing_extensions.Literal["author_date_offset"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_committer",b"_committer"]) -> typing.Optional[typing_extensions.Literal["committer"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_committer_date",b"_committer_date"]) -> typing.Optional[typing_extensions.Literal["committer_date"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_committer_date_offset",b"_committer_date_offset"]) -> typing.Optional[typing_extensions.Literal["committer_date_offset"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_message",b"_message"]) -> typing.Optional[typing_extensions.Literal["message"]]: ...
global___RevisionData = RevisionData

class ReleaseData(google.protobuf.message.Message):
    """Release node properties"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    AUTHOR_FIELD_NUMBER: builtins.int
    AUTHOR_DATE_FIELD_NUMBER: builtins.int
    AUTHOR_DATE_OFFSET_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    author: builtins.int
    """Release author ID (anonymized)"""

    author_date: builtins.int
    """UNIX timestamp of the release date (UTC)"""

    author_date_offset: builtins.int
    """Timezone of the release author date as an offset from UTC"""

    name: builtins.bytes
    """Release name"""

    message: builtins.bytes
    """Release message"""

    def __init__(self,
        *,
        author: typing.Optional[builtins.int] = ...,
        author_date: typing.Optional[builtins.int] = ...,
        author_date_offset: typing.Optional[builtins.int] = ...,
        name: typing.Optional[builtins.bytes] = ...,
        message: typing.Optional[builtins.bytes] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_author",b"_author","_author_date",b"_author_date","_author_date_offset",b"_author_date_offset","_message",b"_message","_name",b"_name","author",b"author","author_date",b"author_date","author_date_offset",b"author_date_offset","message",b"message","name",b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_author",b"_author","_author_date",b"_author_date","_author_date_offset",b"_author_date_offset","_message",b"_message","_name",b"_name","author",b"author","author_date",b"author_date","author_date_offset",b"author_date_offset","message",b"message","name",b"name"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author",b"_author"]) -> typing.Optional[typing_extensions.Literal["author"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author_date",b"_author_date"]) -> typing.Optional[typing_extensions.Literal["author_date"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_author_date_offset",b"_author_date_offset"]) -> typing.Optional[typing_extensions.Literal["author_date_offset"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_message",b"_message"]) -> typing.Optional[typing_extensions.Literal["message"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_name",b"_name"]) -> typing.Optional[typing_extensions.Literal["name"]]: ...
global___ReleaseData = ReleaseData

class OriginData(google.protobuf.message.Message):
    """Origin node properties"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    URL_FIELD_NUMBER: builtins.int
    url: typing.Text
    """URL of the origin"""

    def __init__(self,
        *,
        url: typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_url",b"_url","url",b"url"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_url",b"_url","url",b"url"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_url",b"_url"]) -> typing.Optional[typing_extensions.Literal["url"]]: ...
global___OriginData = OriginData

class EdgeLabel(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    PERMISSION_FIELD_NUMBER: builtins.int
    VISIT_TIMESTAMP_FIELD_NUMBER: builtins.int
    IS_FULL_VISIT_FIELD_NUMBER: builtins.int
    name: builtins.bytes
    """Directory entry name for directories, branch name for snapshots"""

    permission: builtins.int
    """Entry permission (only set for directories)."""

    visit_timestamp: builtins.int
    """For origin->snapshot (or snapshot->origin in the transposed graph), this is
    the UNIX timestamp (UTC) of the visit that found the snapshot to be
    the current state of the origin at that time.
    """

    is_full_visit: builtins.bool
    """For origin->snapshot (or snapshot->origin in the transposed graph), this
    indicates whether the visit was fully complete; ie. if the snapshot is the
    full state of the origin (instead of a partial state).
    """

    def __init__(self,
        *,
        name: typing.Optional[builtins.bytes] = ...,
        permission: typing.Optional[builtins.int] = ...,
        visit_timestamp: typing.Optional[builtins.int] = ...,
        is_full_visit: typing.Optional[builtins.bool] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_is_full_visit",b"_is_full_visit","_name",b"_name","_permission",b"_permission","_visit_timestamp",b"_visit_timestamp","is_full_visit",b"is_full_visit","name",b"name","permission",b"permission","visit_timestamp",b"visit_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_is_full_visit",b"_is_full_visit","_name",b"_name","_permission",b"_permission","_visit_timestamp",b"_visit_timestamp","is_full_visit",b"is_full_visit","name",b"name","permission",b"permission","visit_timestamp",b"visit_timestamp"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_is_full_visit",b"_is_full_visit"]) -> typing.Optional[typing_extensions.Literal["is_full_visit"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_name",b"_name"]) -> typing.Optional[typing_extensions.Literal["name"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_permission",b"_permission"]) -> typing.Optional[typing_extensions.Literal["permission"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_visit_timestamp",b"_visit_timestamp"]) -> typing.Optional[typing_extensions.Literal["visit_timestamp"]]: ...
global___EdgeLabel = EdgeLabel

class CountResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    COUNT_FIELD_NUMBER: builtins.int
    count: builtins.int
    def __init__(self,
        *,
        count: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["count",b"count"]) -> None: ...
global___CountResponse = CountResponse

class StatsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___StatsRequest = StatsRequest

class StatsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class NumNodesByTypeEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: builtins.int
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class NumArcsByTypeEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: builtins.int
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    NUM_NODES_FIELD_NUMBER: builtins.int
    NUM_EDGES_FIELD_NUMBER: builtins.int
    COMPRESSION_RATIO_FIELD_NUMBER: builtins.int
    BITS_PER_NODE_FIELD_NUMBER: builtins.int
    BITS_PER_EDGE_FIELD_NUMBER: builtins.int
    AVG_LOCALITY_FIELD_NUMBER: builtins.int
    INDEGREE_MIN_FIELD_NUMBER: builtins.int
    INDEGREE_MAX_FIELD_NUMBER: builtins.int
    INDEGREE_AVG_FIELD_NUMBER: builtins.int
    OUTDEGREE_MIN_FIELD_NUMBER: builtins.int
    OUTDEGREE_MAX_FIELD_NUMBER: builtins.int
    OUTDEGREE_AVG_FIELD_NUMBER: builtins.int
    EXPORT_STARTED_AT_FIELD_NUMBER: builtins.int
    EXPORT_ENDED_AT_FIELD_NUMBER: builtins.int
    NUM_NODES_BY_TYPE_FIELD_NUMBER: builtins.int
    NUM_ARCS_BY_TYPE_FIELD_NUMBER: builtins.int
    num_nodes: builtins.int
    """Number of nodes in the graph"""

    num_edges: builtins.int
    """Number of edges in the graph"""

    compression_ratio: builtins.float
    """Ratio between the graph size and the information-theoretical lower
    bound
    """

    bits_per_node: builtins.float
    """Number of bits per node (overall graph size in bits divided by the
    number of nodes)
    """

    bits_per_edge: builtins.float
    """Number of bits per edge (overall graph size in bits divided by the
    number of arcs).
    """

    avg_locality: builtins.float
    indegree_min: builtins.int
    """Smallest indegree"""

    indegree_max: builtins.int
    """Largest indegree"""

    indegree_avg: builtins.float
    """Average indegree"""

    outdegree_min: builtins.int
    """Smallest outdegree"""

    outdegree_max: builtins.int
    """Largest outdegree"""

    outdegree_avg: builtins.float
    """Average outdegree"""

    export_started_at: builtins.int
    """Time when the export started"""

    export_ended_at: builtins.int
    """Time when the export ended"""

    @property
    def num_nodes_by_type(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, builtins.int]:
        """maps node types (ie. 'cnt', 'dir', etc.) to the number of nodes of that type"""
        pass
    @property
    def num_arcs_by_type(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, builtins.int]:
        """maps edge types (ie. 'dir:cnt', 'dir:dir', 'rev:'dir', etc.) to the number of nodes of that type"""
        pass
    def __init__(self,
        *,
        num_nodes: builtins.int = ...,
        num_edges: builtins.int = ...,
        compression_ratio: typing.Optional[builtins.float] = ...,
        bits_per_node: typing.Optional[builtins.float] = ...,
        bits_per_edge: typing.Optional[builtins.float] = ...,
        avg_locality: typing.Optional[builtins.float] = ...,
        indegree_min: builtins.int = ...,
        indegree_max: builtins.int = ...,
        indegree_avg: builtins.float = ...,
        outdegree_min: builtins.int = ...,
        outdegree_max: builtins.int = ...,
        outdegree_avg: builtins.float = ...,
        export_started_at: typing.Optional[builtins.int] = ...,
        export_ended_at: typing.Optional[builtins.int] = ...,
        num_nodes_by_type: typing.Optional[typing.Mapping[typing.Text, builtins.int]] = ...,
        num_arcs_by_type: typing.Optional[typing.Mapping[typing.Text, builtins.int]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_avg_locality",b"_avg_locality","_bits_per_edge",b"_bits_per_edge","_bits_per_node",b"_bits_per_node","_compression_ratio",b"_compression_ratio","_export_ended_at",b"_export_ended_at","_export_started_at",b"_export_started_at","avg_locality",b"avg_locality","bits_per_edge",b"bits_per_edge","bits_per_node",b"bits_per_node","compression_ratio",b"compression_ratio","export_ended_at",b"export_ended_at","export_started_at",b"export_started_at"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_avg_locality",b"_avg_locality","_bits_per_edge",b"_bits_per_edge","_bits_per_node",b"_bits_per_node","_compression_ratio",b"_compression_ratio","_export_ended_at",b"_export_ended_at","_export_started_at",b"_export_started_at","avg_locality",b"avg_locality","bits_per_edge",b"bits_per_edge","bits_per_node",b"bits_per_node","compression_ratio",b"compression_ratio","export_ended_at",b"export_ended_at","export_started_at",b"export_started_at","indegree_avg",b"indegree_avg","indegree_max",b"indegree_max","indegree_min",b"indegree_min","num_arcs_by_type",b"num_arcs_by_type","num_edges",b"num_edges","num_nodes",b"num_nodes","num_nodes_by_type",b"num_nodes_by_type","outdegree_avg",b"outdegree_avg","outdegree_max",b"outdegree_max","outdegree_min",b"outdegree_min"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_avg_locality",b"_avg_locality"]) -> typing.Optional[typing_extensions.Literal["avg_locality"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_bits_per_edge",b"_bits_per_edge"]) -> typing.Optional[typing_extensions.Literal["bits_per_edge"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_bits_per_node",b"_bits_per_node"]) -> typing.Optional[typing_extensions.Literal["bits_per_node"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_compression_ratio",b"_compression_ratio"]) -> typing.Optional[typing_extensions.Literal["compression_ratio"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_export_ended_at",b"_export_ended_at"]) -> typing.Optional[typing_extensions.Literal["export_ended_at"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_export_started_at",b"_export_started_at"]) -> typing.Optional[typing_extensions.Literal["export_started_at"]]: ...
global___StatsResponse = StatsResponse



# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module defines the various objects of the
:ref:`swh.graph example dataset <swh-graph-example-dataset>`.
"""

import datetime
from pathlib import Path
from typing import List, Union

from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    Release,
    Revision,
    RevisionType,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    Timestamp,
    TimestampWithTimezone,
)


def h(id: int, width=40) -> bytes:
    """
    :meta private:
    """
    return bytes.fromhex(f"{id:0{width}}")


PERSONS: List[Person] = [
    Person(fullname=b"foo", name=b"foo", email=b""),
    Person(fullname=b"bar", name=b"bar", email=b""),
    Person(fullname=b"baz", name=b"baz", email=b""),
]
"""Example :py:class:`swh.model.model.Person` instances

  Each Person is respectively named “foo“, “bar”, and “baz”.
  They are used as authors in :py:const:`REVISIONS` and :py:const:`RELEASES`.

  :meta hide-value:
"""  # pylint: disable=W0105

CONTENTS: List[Content] = [
    Content(sha1_git=h(1), sha1=h(1), sha256=h(1, 64), blake2s256=h(1, 64), length=42),
    Content(sha1_git=h(4), sha1=h(4), sha256=h(4, 64), blake2s256=h(4, 64), length=404),
    Content(
        sha1_git=h(5), sha1=h(5), sha256=h(5, 64), blake2s256=h(5, 64), length=1337
    ),
    Content(sha1_git=h(7), sha1=h(7), sha256=h(7, 64), blake2s256=h(7, 64), length=666),
    Content(
        sha1_git=h(11), sha1=h(11), sha256=h(11, 64), blake2s256=h(11, 64), length=313
    ),
    Content(
        sha1_git=h(14), sha1=h(14), sha256=h(14, 64), blake2s256=h(14, 64), length=14
    ),
]
"""Example :py:class:`swh.model.model.Content` instances

  :meta hide-value:

  Objects have the following SWHIDs:

  - ``swh:1:cnt:0000000000000000000000000000000000000001``
  - ``swh:1:cnt:0000000000000000000000000000000000000004``
  - ``swh:1:cnt:0000000000000000000000000000000000000005``
  - ``swh:1:cnt:0000000000000000000000000000000000000007``
  - ``swh:1:cnt:0000000000000000000000000000000000000011``
  - ``swh:1:cnt:0000000000000000000000000000000000000014``
"""  # pylint: disable=W0105
# "\n".join([f"  - {c.swhid()}" for c in CONTENTS])


SKIPPED_CONTENTS: List[SkippedContent] = [
    SkippedContent(
        sha1_git=h(15),
        sha1=h(15),
        sha256=h(15, 64),
        blake2s256=h(15, 64),
        length=404,
        status="absent",
        reason="Not found",
    ),
]
"""Example :py:class:`swh.model.model.SkippedContent` instances

  :meta hide-value:

  Object has the following SWHIDs:

  - ``swh:1:cnt:0000000000000000000000000000000000000015``
"""  # pylint: disable=W0105
# "\n".join([f"  - {c.swhid()}" for c in SKIPPED_CONTENTS])


DIRECTORIES: List[Directory] = [
    Directory(
        id=h(2),
        entries=(
            DirectoryEntry(
                name=b"README.md",
                perms=0o100644,
                type="file",
                target=h(1),
            ),
        ),
    ),
    Directory(
        id=h(6),
        entries=(
            DirectoryEntry(
                name=b"README.rst",
                perms=0o100644,
                type="file",
                target=h(4),
            ),
            DirectoryEntry(
                name=b"main.py",
                perms=0o100644,
                type="file",
                target=h(5),
            ),
        ),
    ),
    Directory(
        id=h(8),
        entries=(
            DirectoryEntry(
                name=b"README.md",
                perms=0o100644,
                type="file",
                target=h(1),
            ),
            DirectoryEntry(
                name=b"parser.c",
                perms=0o100644,
                type="file",
                target=h(7),
            ),
            DirectoryEntry(
                name=b"tests",
                perms=0o100755,
                type="dir",
                target=h(6),
            ),
        ),
    ),
    Directory(
        id=h(12),
        entries=(
            DirectoryEntry(
                name=b"README.md",
                perms=0o100644,
                type="file",
                target=h(11),
            ),
            DirectoryEntry(
                name=b"oldproject",
                perms=0o100755,
                type="dir",
                target=h(8),
            ),
        ),
    ),
    Directory(
        id=h(16),
        entries=(
            DirectoryEntry(
                name=b"TODO.txt",
                perms=0o100644,
                type="file",
                target=h(15),
            ),
        ),
    ),
    Directory(
        id=h(17),
        entries=(
            DirectoryEntry(
                name=b"TODO.txt",
                perms=0o100644,
                type="file",
                target=h(14),
            ),
            DirectoryEntry(
                name=b"old",
                perms=0o100755,
                type="dir",
                target=h(16),
            ),
        ),
    ),
]
""" Example :py:class:`swh.model.model.Directory` instances

  Directories have the following SWHIDs and content:

  * ``swh:1:dir:0000000000000000000000000000000000000002``

    - ``README.md`` → ``swh:1:cnt:0000000000000000000000000000000000000001``

  * ``swh:1:dir:0000000000000000000000000000000000000006``

    - ``README.rst`` → ``swh:1:cnt:0000000000000000000000000000000000000004``
    - ``main.py`` → ``swh:1:cnt:0000000000000000000000000000000000000005``

  * ``swh:1:dir:0000000000000000000000000000000000000008``

    - ``README.md`` → ``swh:1:cnt:0000000000000000000000000000000000000001``
    - ``parser.c`` → ``swh:1:cnt:0000000000000000000000000000000000000007``
    - ``tests`` → ``swh:1:dir:0000000000000000000000000000000000000006``

  * ``swh:1:dir:0000000000000000000000000000000000000012``

    - ``README.md`` → ``swh:1:cnt:0000000000000000000000000000000000000011``
    - ``oldproject`` → ``swh:1:dir:0000000000000000000000000000000000000008``

  * ``swh:1:dir:0000000000000000000000000000000000000016``

    - ``TODO.txt`` → ``swh:1:cnt:0000000000000000000000000000000000000015``

  * ``swh:1:dir:0000000000000000000000000000000000000017``

    - ``TODO.txt`` → ``swh:1:cnt:0000000000000000000000000000000000000014``
    - ``old`` → ``swh:1:dir:0000000000000000000000000000000000000016``

  :meta hide-value:
"""  # pylint: disable=W0105
# "\n".join(
#     itertools.chain(
#         *[
#             [
#                 f"  * ``{d.swhid()}``",
#                 "",
#                 *[
#                     f"    - ``{e.name.decode('us-ascii')}`` → ``{e.swhid()}``"
#                     for e in d.entries
#                 ],
#                 "",
#             ]
#             for d in DIRECTORIES
#         ]
#     )
# )


REVISIONS: List[Revision] = [
    Revision(
        id=h(3),
        message=b"Initial commit",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111122220,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        committer=PERSONS[0],
        author=PERSONS[0],
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111122220,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        type=RevisionType.GIT,
        directory=h(2),
        synthetic=False,
        metadata=None,
        parents=(),
    ),
    Revision(
        id=h(9),
        message=b"Add parser",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111144440,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        committer=PERSONS[1],
        author=PERSONS[1],
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111155550,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        type=RevisionType.GIT,
        directory=h(8),
        synthetic=False,
        metadata=None,
        parents=(h(3),),
    ),
    Revision(
        id=h(13),
        message=b"Add tests",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111166660,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        committer=PERSONS[1],
        author=PERSONS[0],
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111166660,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        type=RevisionType.GIT,
        directory=h(12),
        synthetic=False,
        metadata=None,
        parents=(h(9),),
    ),
    Revision(
        id=h(18),
        message=b"Refactor codebase",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111177770,
                microseconds=0,
            ),
            offset_bytes=b"+0000",
        ),
        committer=PERSONS[0],
        author=PERSONS[2],
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1111177770,
                microseconds=0,
            ),
            offset_bytes=b"+0000",
        ),
        type=RevisionType.GIT,
        directory=h(17),
        synthetic=False,
        metadata=None,
        parents=(h(13),),
    ),
]
"""Example :py:class:`swh.model.model.Revision` instances

  :meta hide-value:

  Objects have the following SWHIDs:

  - ``swh:1:rev:0000000000000000000000000000000000000003``
  - ``swh:1:rev:0000000000000000000000000000000000000009``
  - ``swh:1:rev:0000000000000000000000000000000000000013``
  - ``swh:1:rev:0000000000000000000000000000000000000018``
"""  # pylint: disable=W0105
# "\n".join([f"  - {r.swhid()}" for r in REVISIONS])


RELEASES: List[Release] = [
    Release(
        id=h(10),
        name=b"v1.0",
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567890,
                microseconds=0,
            ),
            offset_bytes=b"+0200",
        ),
        author=PERSONS[0],
        target_type=ObjectType.REVISION,
        target=h(9),
        message=b"Version 1.0",
        synthetic=False,
    ),
    Release(
        id=h(19),
        name=b"v2.0",
        date=None,
        author=PERSONS[1],
        target_type=ObjectType.REVISION,
        target=h(18),
        message=b"Version 2.0",
        synthetic=False,
    ),
    Release(
        id=h(21),
        name=b"v2.0-anonymous",
        date=None,
        author=None,
        target_type=ObjectType.REVISION,
        target=h(18),
        message=b"Version 2.0 but with no author",
        synthetic=False,
    ),
]
"""Example :py:class:`swh.model.model.Release` instances

  :meta hide-value:

  Objects have the following SWHIDs:

  - ``swh:1:rel:0000000000000000000000000000000000000010``
  - ``swh:1:rel:0000000000000000000000000000000000000019``
  - ``swh:1:rel:0000000000000000000000000000000000000021``
"""  # pylint: disable=W0105
# "\n".join([f"  - {r.swhid()}" for r in RELEASES])


SNAPSHOTS: List[Snapshot] = [
    Snapshot(
        id=h(20),
        branches={
            b"refs/heads/master": SnapshotBranch(
                target=h(9), target_type=SnapshotTargetType.REVISION
            ),
            b"refs/tags/v1.0": SnapshotBranch(
                target=h(10), target_type=SnapshotTargetType.RELEASE
            ),
        },
    ),
    Snapshot(
        id=h(22),
        branches={
            b"refs/heads/master": SnapshotBranch(
                target=h(9), target_type=SnapshotTargetType.REVISION
            ),
            b"refs/tags/v1.0": SnapshotBranch(
                target=h(10), target_type=SnapshotTargetType.RELEASE
            ),
            b"refs/tags/v2.0-anonymous": SnapshotBranch(
                target=h(21), target_type=SnapshotTargetType.RELEASE
            ),
        },
    ),
]
"""Example :py:class:`swh.model.model.Release` instances

  :meta hide-value:

  Objects have the following SWHIDs and branches:

  * ``swh:1:snp:0000000000000000000000000000000000000020``

    - ``refs/heads/master`` →
      ``swh:1:rev:0000000000000000000000000000000000000009``
    - ``refs/tags/v1.0`` →
      ``swh:1:rel:0000000000000000000000000000000000000010``

  * ``swh:1:snp:0000000000000000000000000000000000000022``

    - ``refs/heads/master`` →
      ``swh:1:rev:0000000000000000000000000000000000000009``
    - ``refs/tags/v1.0`` →
      ``swh:1:rel:0000000000000000000000000000000000000010``
    - ``refs/tags/v2.0-anonymous`` →
      ``swh:1:rel:0000000000000000000000000000000000000021``
"""  # pylint: disable=W0105
# "\n".join(
#     itertools.chain(
#         *[
#             [
#                 "",
#                 f"  * ``{d.swhid()}``",
#                 "",
#                 *[f"    - ``{name.decode('us-ascii')}`` →\n"
#                    "      ``{b.swhid()}``"
#                    for name, b in d.branches.items()],
#             ]
#             for d in SNAPSHOTS
#         ]
#     )
# )

ORIGIN_VISITS: List[OriginVisit] = [
    OriginVisit(
        origin="https://example.com/swh/graph",
        date=datetime.datetime(
            2013, 5, 7, 4, 20, 39, 369271, tzinfo=datetime.timezone.utc
        ),
        visit=1,
        type="git",
    ),
    OriginVisit(
        origin="https://example.com/swh/graph2",
        date=datetime.datetime(
            2013, 5, 7, 4, 20, 39, 369271, tzinfo=datetime.timezone.utc
        ),
        visit=1,
        type="git",
    ),
]
"""Example :py:class:`swh.model.model.OriginVisit` instances

  :meta hide-value:

  Objects have the following origins:

  - ``https://example.com/swh/graph`` (``swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054``)
  - ``https://example.com/swh/graph2`` (``swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165``)
"""  # pylint: disable=W0105


ORIGIN_VISIT_STATUSES: List[OriginVisitStatus] = [
    OriginVisitStatus(
        origin="https://example.com/swh/graph",
        date=datetime.datetime(
            2013, 5, 7, 4, 20, 41, 369271, tzinfo=datetime.timezone.utc
        ),
        visit=1,
        type="git",
        status="full",
        snapshot=h(20),
        metadata=None,
    ),
    OriginVisitStatus(
        origin="https://example.com/swh/graph2",
        date=datetime.datetime(
            2013, 5, 7, 4, 20, 41, 369271, tzinfo=datetime.timezone.utc
        ),
        visit=1,
        type="git",
        status="full",
        snapshot=h(22),
        metadata=None,
    ),
]
"""Example :py:class:`swh.model.model.OriginVisitStatus` instances

  :meta hide-value:

  Objects have the following origins and snapshots:

  - ``swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054`` →
    ``swh:1:snp:0000000000000000000000000000000000000020``
  - ``swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165`` →
    ``swh:1:snp:0000000000000000000000000000000000000022``
"""  # pylint: disable=W0105


INITIAL_ORIGIN: Origin = Origin(url="https://example.com/swh/graph")
"""Origin where the development of the tiny project represented by this
example dataset was initially made.

:meta hide-value:

- URL: ``https://example.com/swh/graph``
- SWHID: ``swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054``
"""  # pylint: disable=W0105


FORKED_ORIGIN: Origin = Origin(url="https://example.com/swh/graph2")
"""Origin that picked up (forked) the development of the tiny project
represented by this example dataset.

:meta hide-value:

- URL: ``https://example.com/swh/graph2``
- SWHID: ``swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165``
"""  # pylint: disable=W0105


ORIGINS: List[Origin] = [INITIAL_ORIGIN, FORKED_ORIGIN]
"""Example :py:class:`swh.model.model.Origin` instances

  :meta hide-value:

  Objects have the following SWHIDs:

  - ``https://example.com/swh/graph`` (``swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054``)
  - ``https://example.com/swh/graph2`` (``swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165``)
"""  # pylint: disable=W0105
# "\n".join([f"  - ``{o.url}`` (``{o.swhid()}``)" for o in ORIGINS]


DATASET: List[
    Union[
        Content,
        SkippedContent,
        Directory,
        Revision,
        Release,
        Snapshot,
        OriginVisit,
        OriginVisitStatus,
        Origin,
    ]
] = [
    *CONTENTS,
    *SKIPPED_CONTENTS,
    *DIRECTORIES,
    *REVISIONS,
    *RELEASES,
    *SNAPSHOTS,
    *ORIGIN_VISITS,
    *ORIGIN_VISIT_STATUSES,
    *ORIGINS,
]
"""Full dataset comprised with all the objects defined above.

  :meta hide-value:
"""  # pylint: disable=W0105


DATASET_DIR: Path = Path(__file__).parent
"""Path to the dataset directory

  :meta hide-value:
"""  # pylint: disable=W0105

SENSITIVE_DIR: Path = Path(f"{DATASET_DIR}-sensitive")
"""Path to the sensitive data directory

  :meta hide-value:
"""



# Copyright (C) 2022-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
from pathlib import Path
from typing import Dict, List, Tuple

import luigi

from swh.export.luigi import AthenaDatabaseTarget

OBJECT_TYPES = {"ori", "snp", "rel", "rev", "dir", "cnt"}


# singleton written to signal to workers they should stop
class _EndOfQueue:
    pass


_ENF_OF_QUEUE = _EndOfQueue()


def estimate_node_count(
    local_graph_path: Path, graph_name: str, object_types: str
) -> int:
    """Returns the number of nodes of the given types (in the 'cnt,dir,rev,rel,snp,ori'
    format) in the graph.

    This is meant to estimate RAM usage, and will return an overapproximation if the
    actual number of nodes is not available yet.
    """
    try:
        return count_nodes(local_graph_path, graph_name, object_types)
    except FileNotFoundError:
        # The graph was not compressed yet, so we can't estimate the memory it will take
        # to run this task, once the compressed graph is available.
        # As a hack, we return a very large value, which will force Luigi to stop before
        # running this, and when starting again (after the compressed graph is available),
        # it will call this function again to get an accurate estimate
        return 10**100


def count_nodes(local_graph_path: Path, graph_name: str, object_types: str) -> int:
    """Returns the number of nodes of the given types (in the 'cnt,dir,rev,rel,snp,ori'
    format) in the graph.
    """
    node_stats = (local_graph_path / f"{graph_name}.nodes.stats.txt").read_text()
    nb_nodes_per_type = dict(line.split() for line in node_stats.split("\n") if line)
    return sum(int(nb_nodes_per_type[type_]) for type_ in object_types.split(","))


class _ParquetToS3Task(luigi.Task):
    """Base class for tasks which take a local Parquet table as input, upload it to S3.

    See :class:`_ParquetToS3ToAthenaTask` for tasks that also need to create an
    Athena table."""

    parallelism = 10

    def _input_parquet_path(self) -> Path:
        raise NotImplementedError(f"{self.__class__.__name__}._input_parquet_path")

    def _s3_bucket(self) -> str:
        raise NotImplementedError(f"{self.__class__.__name__}._s3_bucket")

    def _s3_prefix(self) -> str:
        raise NotImplementedError(f"{self.__class__.__name__}._s3_prefix")

    def _s3_path(self) -> str:
        """Concatenates ``s3://``, :meth:`_s3_bucket`, and :meth:`_s3_prefix`."""
        return f"s3://{self._s3_bucket()}/{self._s3_prefix()}"

    def _approx_nb_rows(self) -> int:
        """Returns number of rows in the Parquet file. Used only for progress reporting"""
        import pyarrow.parquet

        return sum(
            pyarrow.parquet.read_metadata(file).num_rows
            for file in self._input_parquet_path().iterdir()
        )

    def output(self) -> luigi.Target:
        import luigi.contrib.s3

        return luigi.contrib.s3.S3FlagTarget(self._s3_path())

    def _upload_files(self):
        """Copies all files in parallel from the path returned by :meth:`_input_parquet_path`
        to the S3 bucket and path returned by :meth:`_s3_path`."""
        import multiprocessing

        import tqdm

        self.__status_messages: Dict[Path, str] = {}

        paths = list(self._input_parquet_path().glob("**/*.parquet"))

        with multiprocessing.Pool(self.parallelism) as p:
            for i, relative_path in tqdm.tqdm(
                enumerate(p.imap_unordered(self._upload_file, paths)),
                total=len(paths),
                desc=f"Uploading {self._input_parquet_path()} to {self._s3_path()}",
            ):
                self.set_progress_percentage(int(i * 100 / len(paths)))
                self.set_status_message("\n".join(self.__status_messages.values()))

    def _upload_file(self, path):
        import luigi.contrib.s3

        client = luigi.contrib.s3.S3Client()

        relative_path = path.relative_to(self._input_parquet_path())

        self.__status_messages[path] = f"Uploading {relative_path}"

        client.put_multipart(
            path,
            f"{self._s3_path()}/{relative_path}",
            ACL="public-read",
        )

        del self.__status_messages[path]

        return relative_path

    def run(self) -> None:
        """Calls :meth:`_upload_files`, then writes a :file:`_SUCCESS` stamp."""
        self._upload_files()

        client = luigi.contrib.s3.S3Client()
        client.put_string("success", f"{self._s3_path()}/_SUCCESS")


class _ParquetToS3ToAthenaTask(_ParquetToS3Task):
    def _parquet_columns(self) -> List[Tuple[str, str]]:
        """Returns a list of ``(column_name, parquet_type)``"""
        raise NotImplementedError(f"{self.__class__.__name__}._parquet_columns")

    def _athena_db_name(self) -> str:
        raise NotImplementedError(f"{self.__class__.__name__}._athena_db_name")

    def _athena_table_name(self) -> str:
        raise NotImplementedError(f"{self.__class__.__name__}._athena_table_name")

    def output(self) -> luigi.Target:
        return AthenaDatabaseTarget(self._athena_db_name(), {self._athena_table_name()})

    def run(self) -> None:
        """Calls :meth:`_upload_files` then :meth:`_create_athena_table`"""
        self._upload_files()

        self._create_athena_table()

    def _create_athena_table(self):
        import boto3

        from swh.export.athena import query

        client = boto3.client("athena")
        client.output_location = self.s3_athena_output_location

        client.database_name = "default"  # we have to pick some existing database
        query(
            client,
            f"CREATE DATABASE IF NOT EXISTS {self._athena_db_name()};",
            desc=f"Creating {self._athena_db_name()} database",
        )
        client.database_name = self._athena_db_name()

        columns = ", ".join(
            f"{col} {type_}" for (col, type_) in self._parquet_columns()
        )

        query(
            client,
            f"""
            CREATE EXTERNAL TABLE IF NOT EXISTS
            {self._athena_db_name()}.{self._athena_table_name()}
            ({columns})
            {self.create_table_extras()}
            STORED AS PARQUET
            LOCATION 's3://{self._s3_bucket()}/{self._s3_prefix()}';
            """,
            desc=f"Creating table {self._athena_table_name()}",
        )

        # needed for partitioned tables
        query(
            client,
            f"MSCK REPAIR TABLE `{self._athena_table_name()}`",
            desc=f"'Repairing' table {self._athena_table_name()}",
        )

    def create_table_extras(self) -> str:
        """Extra clauses to add to the ``CREATE EXTERNAL TABLE`` statement."""
        return ""



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging

from swh.core.api import RPCClient

logger = logging.getLogger(__name__)


class GraphAPIError(Exception):
    """Graph API Error"""

    def __str__(self):
        return """An unexpected error occurred
               in the Graph backend: {}""".format(
            self.args
        )


class GraphArgumentException(Exception):
    def __init__(self, *args, response=None):
        super().__init__(*args)
        self.response = response


class RemoteGraphClient(RPCClient):
    """Client to the Software Heritage Graph."""

    def __init__(self, url, timeout=None):
        super().__init__(api_exception=GraphAPIError, url=url, timeout=timeout)
        try:
            stats = self.stats()
        except GraphArgumentException as e:
            if e.response.status_code == 404:
                raise ValueError(
                    "URL is incorrect (got 404 while trying to retrieve stats)"
                ) from None
            raise
        if "num_nodes" not in stats:
            raise ValueError("stats returned unexpected results (no `num_nodes` entry)")
        if "export_started_at" in stats:
            from datetime import datetime, timezone

            logger.debug(
                "Graph export started at %s (%d nodes)",
                datetime.fromtimestamp(
                    int(stats["export_started_at"]), tz=timezone.utc
                ).isoformat(),
                stats["num_nodes"],
            )

    def raw_verb_lines(self, verb, endpoint, **kwargs):
        response = self.raw_verb(verb, endpoint, stream=True, **kwargs)
        self.raise_for_status(response)
        for line in response.iter_lines():
            content = line.decode().lstrip("\n")
            if content:
                yield content

    def get_lines(self, endpoint, **kwargs):
        yield from self.raw_verb_lines("get", endpoint, **kwargs)

    def raise_for_status(self, response) -> None:
        if response.status_code // 100 == 4:
            raise GraphArgumentException(
                response.content.decode("ascii"), response=response
            )
        super().raise_for_status(response)

    # Web API endpoints

    def stats(self):
        return self._get("stats")

    def leaves(
        self,
        src,
        edges="*",
        direction="forward",
        max_edges=0,
        return_types="*",
        max_matching_nodes=0,
    ):
        return self.get_lines(
            "leaves/{}".format(src),
            params={
                "edges": edges,
                "direction": direction,
                "max_edges": max_edges,
                "return_types": return_types,
                "max_matching_nodes": max_matching_nodes,
            },
        )

    def neighbors(
        self,
        src,
        edges="*",
        direction="forward",
        max_edges=0,
        return_types="*",
        max_matching_nodes=0,
    ):
        return self.get_lines(
            "neighbors/{}".format(src),
            params={
                "edges": edges,
                "direction": direction,
                "max_edges": max_edges,
                "return_types": return_types,
                "max_matching_nodes": max_matching_nodes,
            },
        )

    def visit_nodes(
        self,
        src,
        edges="*",
        direction="forward",
        max_edges=0,
        return_types="*",
        max_matching_nodes=0,
    ):
        return self.get_lines(
            "visit/nodes/{}".format(src),
            params={
                "edges": edges,
                "direction": direction,
                "max_edges": max_edges,
                "return_types": return_types,
                "max_matching_nodes": max_matching_nodes,
            },
        )

    def visit_edges(self, src, edges="*", direction="forward", max_edges=0):
        for edge in self.get_lines(
            "visit/edges/{}".format(src),
            params={"edges": edges, "direction": direction, "max_edges": max_edges},
        ):
            yield tuple(edge.split())

    def visit_paths(self, src, edges="*", direction="forward", max_edges=0):
        def decode_path_wrapper(it):
            for e in it:
                yield json.loads(e)

        return decode_path_wrapper(
            self.get_lines(
                "visit/paths/{}".format(src),
                params={"edges": edges, "direction": direction, "max_edges": max_edges},
            )
        )

    def walk(
        self, src, dst, edges="*", traversal="dfs", direction="forward", limit=None
    ):
        endpoint = "walk/{}/{}"
        return self.get_lines(
            endpoint.format(src, dst),
            params={
                "edges": edges,
                "traversal": traversal,
                "direction": direction,
                "limit": limit,
            },
        )

    def random_walk(
        self, src, dst, edges="*", direction="forward", limit=None, return_types="*"
    ):
        endpoint = "randomwalk/{}/{}"
        return self.get_lines(
            endpoint.format(src, dst),
            params={
                "edges": edges,
                "direction": direction,
                "limit": limit,
                "return_types": return_types,
            },
        )

    def count_leaves(self, src, edges="*", direction="forward", max_matching_nodes=0):
        return self._get(
            "leaves/count/{}".format(src),
            params={
                "edges": edges,
                "direction": direction,
                "max_matching_nodes": max_matching_nodes,
            },
        )

    def count_neighbors(self, src, edges="*", direction="forward"):
        return self._get(
            "neighbors/count/{}".format(src),
            params={"edges": edges, "direction": direction},
        )

    def count_visit_nodes(
        self, src, edges="*", direction="forward", max_matching_nodes=0
    ):
        return self._get(
            "visit/nodes/count/{}".format(src),
            params={
                "edges": edges,
                "direction": direction,
                "max_matching_nodes": max_matching_nodes,
            },
        )



# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""WebGraph driver"""

from datetime import datetime
from enum import Enum
import logging
import os
from pathlib import Path
import shlex
import subprocess
import sys
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
from swh.graph.config import check_config_compress

if TYPE_CHECKING:
    from .shell import RunResult

logger = logging.getLogger(__name__)


class CompressionSubprocessError(Exception):
    def __init__(self, message: str, log_path: Path):
        super().__init__(f"{message}; full logs at {log_path}")
        self.message = message
        self.log_path = log_path


class CompressionStep(Enum):
    EXTRACT_NODES = -20
    EXTRACT_LABELS = -10
    NODE_STATS = 0
    EDGE_STATS = 3
    LABEL_STATS = 6
    MPH = 10
    BV = 30
    BV_EF = 40
    BFS_ROOTS = 50
    BFS = 60
    PERMUTE_AND_SIMPLIFY_BFS = 70
    BFS_EF = 80
    BFS_DCF = 90
    LLP = 100
    COMPOSE_ORDERS = 110
    PERMUTE_LLP = 120
    EF = 140
    TRANSPOSE = 160
    TRANSPOSE_EF = 170
    MAPS = 180
    EXTRACT_PERSONS = 190
    PERSONS_STATS = 195
    MPH_PERSONS = 200
    EXTRACT_FULLNAMES = 203
    FULLNAMES_EF = 207
    NODE_PROPERTIES = 210
    MPH_LABELS = 220
    LABELS_ORDER = 225
    FCL_LABELS = 230
    EDGE_LABELS = 240
    EDGE_LABELS_TRANSPOSE = 250
    EDGE_LABELS_EF = 270
    EDGE_LABELS_TRANSPOSE_EF = 280
    STATS = 290
    E2E_TEST = 295
    CLEAN_TMP = 300

    def __str__(self):
        return self.name


# full compression pipeline
COMP_SEQ = list(CompressionStep)

# Mapping from compression steps to shell commands implementing them. Commands
# will be executed by the shell, so be careful with meta characters. They are
# specified here as lists of tokens that will be joined together only for ease
# of line splitting. In commands, {tokens} will be interpolated with
# configuration values, see :func:`compress`.
STEP_ARGV: Dict[CompressionStep, List[str]] = {
    CompressionStep.EXTRACT_NODES: [
        "{rust_executable_dir}/swh-graph-extract",
        "extract-nodes",
        "--format",
        "orc",
        "--allowed-node-types",
        "{object_types}",
        "{in_dir}",
        "{out_dir}/{graph_name}.nodes/",
    ],
    CompressionStep.EXTRACT_LABELS: [
        "{rust_executable_dir}/swh-graph-extract",
        "extract-labels",
        "--format",
        "orc",
        "--allowed-node-types",
        "{object_types}",
        "{in_dir}",
        "| zstdmt > {out_dir}/{graph_name}.labels.csv.zst",
    ],
    CompressionStep.NODE_STATS: [
        "{rust_executable_dir}/swh-graph-extract",
        "node-stats",
        "--format",
        "orc",
        "--swhids-dir",
        "{out_dir}/{graph_name}.nodes/",
        "--target-stats",
        "{out_dir}/{graph_name}.nodes.stats.txt",
        "--target-count",
        "{out_dir}/{graph_name}.nodes.count.txt",
    ],
    CompressionStep.EDGE_STATS: [
        "{rust_executable_dir}/swh-graph-extract",
        "edge-stats",
        "--format",
        "orc",
        "--allowed-node-types",
        "{object_types}",
        "--dataset-dir",
        "{in_dir}",
        "--target-stats",
        "{out_dir}/{graph_name}.edges.stats.txt",
        "--target-count",
        "{out_dir}/{graph_name}.edges.count.txt",
    ],
    CompressionStep.LABEL_STATS: [
        "zstdcat {out_dir}/{graph_name}.labels.csv.zst "
        "| wc -l "
        "> {out_dir}/{graph_name}.labels.count.txt"
    ],
    CompressionStep.MPH: [
        "{rust_executable_dir}/swh-graph-compress",
        "pthash-swhids",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "{out_dir}/{graph_name}.nodes/",
        "{out_dir}/{graph_name}.pthash",
    ],
    CompressionStep.BV: [
        "{rust_executable_dir}/swh-graph-extract",
        "bv",
        "--allowed-node-types",
        "{object_types}",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "{in_dir}",
        "{out_dir}/{graph_name}-base",
    ],
    CompressionStep.BV_EF: [
        "{rust_executable_dir}/swh-graph-index",
        "ef",
        "{out_dir}/{graph_name}-base",
    ],
    CompressionStep.BFS_ROOTS: [
        "{rust_executable_dir}/swh-graph-extract",
        "bfs-roots",
        "{in_dir}",
        "{out_dir}/{graph_name}-bfs.roots.txt",
    ],
    CompressionStep.BFS: [
        "{rust_executable_dir}/swh-graph-compress",
        "bfs",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}.pthash",
        "--init-roots",
        "{out_dir}/{graph_name}-bfs.roots.txt",
        "{out_dir}/{graph_name}-base",
        "{out_dir}/{graph_name}-bfs.order",
    ],
    CompressionStep.PERMUTE_AND_SIMPLIFY_BFS: [
        "{rust_executable_dir}/swh-graph-compress",
        "permute-and-symmetrize",
        "{out_dir}/{graph_name}-base",
        "{out_dir}/{graph_name}-bfs-simplified",
        "--permutation",
        "{out_dir}/{graph_name}-bfs.order",
    ],
    CompressionStep.BFS_EF: [
        "{rust_executable_dir}/swh-graph-index",
        "ef",
        "{out_dir}/{graph_name}-bfs-simplified",
    ],
    CompressionStep.BFS_DCF: [
        "{rust_executable_dir}/swh-graph-index",
        "dcf",
        "{out_dir}/{graph_name}-bfs-simplified",
    ],
    CompressionStep.LLP: [
        "{rust_executable_dir}/swh-graph-compress",
        "llp",
        "-g",
        "{llp_gammas}",
        "{out_dir}/{graph_name}-bfs-simplified",
        "{out_dir}/{graph_name}-llp.order",
    ],
    CompressionStep.COMPOSE_ORDERS: [
        "{rust_executable_dir}/swh-graph-compress",
        "compose-orders",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "--input",
        "{out_dir}/{graph_name}-bfs.order",
        "--input",
        "{out_dir}/{graph_name}-llp.order",
        "--output",
        "{out_dir}/{graph_name}.pthash.order",
    ],
    CompressionStep.PERMUTE_LLP: [
        "{rust_executable_dir}/swh-graph-compress",
        "permute",
        "{out_dir}/{graph_name}-base",
        "{out_dir}/{graph_name}",
        "--permutation",
        "{out_dir}/{graph_name}.pthash.order",
    ],
    CompressionStep.EF: [
        "{rust_executable_dir}/swh-graph-index",
        "ef",
        "{out_dir}/{graph_name}",
    ],
    CompressionStep.TRANSPOSE: [
        "{rust_executable_dir}/swh-graph-compress",
        "transpose",
        "{out_dir}/{graph_name}",
        "{out_dir}/{graph_name}-transposed",
    ],
    CompressionStep.TRANSPOSE_EF: [
        "{rust_executable_dir}/swh-graph-index",
        "ef",
        "{out_dir}/{graph_name}-transposed",
    ],
    CompressionStep.MAPS: [
        "{rust_executable_dir}/swh-graph-compress",
        "maps",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "--swhids-dir",
        "{out_dir}/{graph_name}.nodes/",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}.pthash",
        "--order",
        "{out_dir}/{graph_name}.pthash.order",
        "--node2swhid",
        "{out_dir}/{graph_name}.node2swhid.bin",
        "--node2type",
        "{out_dir}/{graph_name}.node2type.bin",
    ],
    CompressionStep.EXTRACT_PERSONS: [
        "{rust_executable_dir}/swh-graph-extract",
        "extract-persons",
        "--allowed-node-types",
        "{object_types}",
        "{in_dir}",
        "| zstdmt > {out_dir}/{graph_name}.persons.csv.zst",
    ],
    CompressionStep.MPH_PERSONS: [
        # skip this step when compressing a graph with no revisions or releases
        'if [[ $(cat {out_dir}/{graph_name}.persons.count.txt) != "0" ]]; then\n',
        "{rust_executable_dir}/swh-graph-compress",
        "pthash-persons",
        "--num-persons",
        "$(cat {out_dir}/{graph_name}.persons.count.txt)",
        "<(zstdcat {out_dir}/{graph_name}.persons.csv.zst)",
        "{out_dir}/{graph_name}.persons.pthash\n",
        "else\n",
        "echo '' > {out_dir}/{graph_name}.persons.pthash;",
        "fi",
    ],
    CompressionStep.EXTRACT_FULLNAMES: [
        # skip this step when compressing a graph with no revisions or releases
        'if [[ $(cat {out_dir}/{graph_name}.persons.count.txt) != "0" ]] &&',
        '[ -d "{in_dir}/person" ]; then\n',
        "{rust_executable_dir}/swh-graph-extract",
        "extract-fullnames",
        "--person-function",
        "{out_dir}/{graph_name}.persons.pthash",
        "{in_dir}",
        "{out_dir}/{graph_name}.persons",
        "{out_dir}/{graph_name}.persons.lengths\n",
        "fi",
    ],
    CompressionStep.FULLNAMES_EF: [
        # skip this step when compressing a graph with no revisions or releases
        'if [[ $(cat {out_dir}/{graph_name}.persons.count.txt) != "0" ]] &&',
        '[ -d "{in_dir}/person" ]; then\n',
        "{rust_executable_dir}/swh-graph-index",
        "fullnames-ef",
        "--num-persons",
        "$(cat {out_dir}/{graph_name}.persons.count.txt)",
        "{out_dir}/{graph_name}.persons",
        "{out_dir}/{graph_name}.persons.lengths",
        "{out_dir}/{graph_name}.persons.ef\n",
        "fi",
    ],
    CompressionStep.PERSONS_STATS: [
        "zstdcat {out_dir}/{graph_name}.persons.csv.zst "
        "| wc -l"
        "> {out_dir}/{graph_name}.persons.count.txt",
    ],
    CompressionStep.NODE_PROPERTIES: [
        "{rust_executable_dir}/swh-graph-extract",
        "node-properties",
        "--format",
        "orc",
        "--allowed-node-types",
        "{object_types}",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}",
        "--order",
        "{out_dir}/{graph_name}.pthash.order",
        "--person-function",
        "{out_dir}/{graph_name}.persons.pthash",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "{in_dir}",
        "{out_dir}/{graph_name}",
    ],
    CompressionStep.MPH_LABELS: [
        "{rust_executable_dir}/swh-graph-compress",
        "pthash-labels",
        "--num-labels",
        "$(cat {out_dir}/{graph_name}.labels.count.txt)",
        "<(zstdcat {out_dir}/{graph_name}.labels.csv.zst)",
        "{out_dir}/{graph_name}.labels.pthash",
    ],
    CompressionStep.LABELS_ORDER: [
        "{rust_executable_dir}/swh-graph-compress",
        "pthash-labels-order",
        "--num-labels",
        "$(cat {out_dir}/{graph_name}.labels.count.txt)",
        "<(zstdcat {out_dir}/{graph_name}.labels.csv.zst)",
        "{out_dir}/{graph_name}.labels.pthash",
        "{out_dir}/{graph_name}.labels.pthash.order",
    ],
    CompressionStep.FCL_LABELS: [
        "{rust_executable_dir}/swh-graph-compress",
        "fcl",
        "--num-lines",
        "$(cat {out_dir}/{graph_name}.labels.count.txt)",
        "<(zstdcat {out_dir}/{graph_name}.labels.csv.zst)",
        "{out_dir}/{graph_name}.labels.fcl",
    ],
    CompressionStep.EDGE_LABELS: [
        "{rust_executable_dir}/swh-graph-extract",
        "edge-labels",
        "--allowed-node-types",
        "{object_types}",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}",
        "--order",
        "{out_dir}/{graph_name}.pthash.order",
        "--label-name-mphf",
        "{out_dir}/{graph_name}.labels.pthash",
        "--label-name-order",
        "{out_dir}/{graph_name}.labels.pthash.order",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "{in_dir}",
        "{out_dir}/{graph_name}",
    ],
    CompressionStep.EDGE_LABELS_TRANSPOSE: [
        "{rust_executable_dir}/swh-graph-extract",
        "edge-labels",
        "--allowed-node-types",
        "{object_types}",
        "--mph-algo",
        "pthash",
        "--function",
        "{out_dir}/{graph_name}",
        "--order",
        "{out_dir}/{graph_name}.pthash.order",
        "--label-name-mphf",
        "{out_dir}/{graph_name}.labels.pthash",
        "--label-name-order",
        "{out_dir}/{graph_name}.labels.pthash.order",
        "--num-nodes",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
        "--transposed",
        "{in_dir}",
        "{out_dir}/{graph_name}-transposed",
    ],
    CompressionStep.EDGE_LABELS_EF: [
        "{rust_executable_dir}/swh-graph-index",
        "labels-ef",
        "{out_dir}/{graph_name}-labelled",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
    ],
    CompressionStep.EDGE_LABELS_TRANSPOSE_EF: [
        "{rust_executable_dir}/swh-graph-index",
        "labels-ef",
        "{out_dir}/{graph_name}-transposed-labelled",
        "$(cat {out_dir}/{graph_name}.nodes.count.txt)",
    ],
    CompressionStep.STATS: [
        "{rust_executable_dir}/swh-graph-compress",
        "stats",
        "--graph",
        "{out_dir}/{graph_name}",
        "--stats",
        "{out_dir}/{graph_name}.stats",
    ],
    CompressionStep.E2E_TEST: [
        sys.executable,
        "-c",
        "'import sys;\
        from swh.graph.e2e_tests import run_e2e_test;\
        run_e2e_test(\
        graph_name=sys.argv[1],\
        in_dir=sys.argv[2],\
        out_dir=sys.argv[3],\
        test_flavor=sys.argv[4],\
        profile=sys.argv[5],\
        )'",
        "{graph_name}",
        "{in_dir}",
        "{out_dir}",
        "{test_flavor}",
        "{profile}",
    ],
    CompressionStep.CLEAN_TMP: [
        "rm",
        "-rf",
        "{out_dir}/{graph_name}-base.*",
        "{out_dir}/{graph_name}-bfs-simplified.*",
        "{out_dir}/{graph_name}-bfs.order",
        "{out_dir}/{graph_name}-llp.order",
        "{out_dir}/{graph_name}.nodes/",
        "{out_dir}/{graph_name}-bfs.roots.txt",
        "{out_dir}/{graph_name}.labels.csv.zst",
        "{out_dir}/{graph_name}.persons.csv.zst",
        "{tmp_dir}",
    ],
}


def do_step(step, conf) -> "List[RunResult]":
    from .shell import Command, CommandException

    log_dir = Path(conf["out_dir"]) / "logs"
    log_dir.mkdir(exist_ok=True)

    step_logger = logger.getChild(f"steps.{step.name.lower()}")
    if not step_logger.isEnabledFor(logging.INFO):
        # Ensure that at least INFO messages are sent, because it is the level we use
        # for the stdout of Rust processes. These processes can take a long time to
        # run, and it would be very annoying to have to run them again just because
        # they crashed with no log.
        step_logger.setLevel(logging.INFO)
    log_path = log_dir / (
        f"{conf['graph_name']}"
        f"-{int(datetime.now().timestamp() * 1000)}"
        f"-{str(step).lower()}.log"
    )
    step_handler = logging.FileHandler(log_path)
    step_logger.addHandler(step_handler)

    step_start_time = datetime.now()
    step_logger.info("Starting compression step %s at %s", step, step_start_time)

    cmd = " ".join(STEP_ARGV[step]).format(
        **{k: shlex.quote(str(v)) for (k, v) in conf.items()}
    )
    cmd_env = os.environ.copy()
    cmd_env["TMPDIR"] = conf["tmp_dir"]
    cmd_env["TZ"] = "UTC"
    cmd_env["RUST_MIN_STACK"] = "8388608"  # 8MiB; avoids stack overflows in LLP
    command = Command.bash(
        "-c",
        cmd,
        env=cmd_env,
        encoding="utf8",
    )._run(stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    step_logger.info("Running: %s", cmd)

    with command.stdout() as stdout:
        for line in stdout:
            step_logger.info(line.rstrip())
    try:
        results = command.wait()
    except CommandException as e:
        msg = f"Compression step {step} returned non-zero exit code {e.returncode}"
        step_logger.critical(msg)
        raise CompressionSubprocessError(msg, log_path)
    step_end_time = datetime.now()
    step_duration = step_end_time - step_start_time
    step_logger.info(
        "Compression step %s finished at %s (in %s)",
        step,
        step_end_time,
        step_duration,
    )
    step_logger.removeHandler(step_handler)
    step_handler.close()
    return results


def compress(
    graph_name: str,
    in_dir: str,
    out_dir: str,
    test_flavor: Optional[str],
    steps: Set[CompressionStep] = set(COMP_SEQ),
    conf: Dict[str, str] = {},
    progress_cb: Callable[[int, CompressionStep], None] = lambda percentage, step: None,
):
    """graph compression pipeline driver from nodes/edges files to compressed
    on-disk representation

    Args:
        graph_name: graph base name, relative to in_dir
        in_dir: input directory, where the uncompressed graph can be found
        out_dir: output directory, where the compressed graph will be stored
        test_flavor: which flavor of tests to run
        steps: compression steps to run (default: all steps)
        conf: compression configuration, supporting the following keys (all are
          optional, so an empty configuration is fine and is the default)

          - batch_size: batch size for `WebGraph transformations
            <http://webgraph.di.unimi.it/docs/it/unimi/dsi/webgraph/Transform.html>`_;
            defaults to 1 billion
          - tmp_dir: temporary directory, defaults to the "tmp" subdir of
            out_dir
          - object_types: comma-separated list of object types to extract
            (eg. ``ori,snp,rel,rev``). Defaults to ``*``.
        progress_cb: a callable taking a percentage and step as argument,
          which is called every time a step starts.

    """
    if not steps:
        steps = set(COMP_SEQ)

    conf = check_config_compress(conf, graph_name, in_dir, out_dir, test_flavor)

    compression_start_time = datetime.now()
    logger.info("Starting compression at %s", compression_start_time)
    seq_no = 0
    for step in COMP_SEQ:
        if step not in steps:
            logger.debug("Skipping compression step %s", step)
            continue
        seq_no += 1
        logger.info("Running compression step %s (%s/%s)", step, seq_no, len(steps))
        progress_cb(seq_no * 100 // len(steps), step)
        do_step(step, conf)
    compression_end_time = datetime.now()
    compression_duration = compression_end_time - compression_start_time
    logger.info("Completed compression in %s", compression_duration)



# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
A proxy HTTP server for swh-graph, talking to the Rust code via the gRPC API.
"""

import json
import logging
import os
from typing import Optional

import aiohttp.test_utils
import aiohttp.web
from google.protobuf import json_format
from google.protobuf.field_mask_pb2 import FieldMask
import grpc

from swh.core.config import read as config_read
from swh.graph.grpc.swhgraph_pb2 import (
    GetNodeRequest,
    NodeFilter,
    StatsRequest,
    TraversalRequest,
)
from swh.graph.grpc.swhgraph_pb2_grpc import TraversalServiceStub
from swh.graph.grpc_server import spawn_rust_grpc_server, stop_grpc_server
from swh.model.swhids import EXTENDED_SWHID_TYPES

try:
    from contextlib import asynccontextmanager
except ImportError:
    # Compatibility with 3.6 backport
    from async_generator import asynccontextmanager  # type: ignore


# maximum number of retries for random walks
RANDOM_RETRIES = 10  # TODO make this configurable via rpc-serve configuration

logger = logging.getLogger(__name__)


async def _aiorpcerror_middleware(app, handler):
    async def middleware_handler(request):
        try:
            return await handler(request)
        except grpc.aio.AioRpcError as e:
            # The default error handler of the RPC framework tries to serialize this
            # with msgpack; which for some unknown reason causes it to raise
            # ValueError("recursion limit exceeded") with a lot of context, causing
            # Sentry to be overflowed with gigabytes of logs (160KB per event, with
            # potentially hundreds of thousands of events per day).
            # Instead, we simply serialize the exception to a string.
            # https://sentry.softwareheritage.org/share/issue/d6d4db971e4b47728a6c1dd06cb9b8a5/
            raise aiohttp.web.HTTPServiceUnavailable(text=str(e))

    return middleware_handler


class GraphServerApp(aiohttp.web.Application):
    def __init__(self, *args, middlewares=(), **kwargs):
        middlewares = (_aiorpcerror_middleware,) + middlewares
        super().__init__(*args, middlewares=middlewares, **kwargs)
        self.on_startup.append(self._start)
        self.on_shutdown.append(self._stop)

    @staticmethod
    async def _start(app):
        app["channel"] = grpc.aio.insecure_channel(app["rpc_url"])
        await app["channel"].__aenter__()
        app["rpc_client"] = TraversalServiceStub(app["channel"])
        await app["rpc_client"].Stats(StatsRequest(), wait_for_ready=True)

    @staticmethod
    async def _stop(app):
        await app["channel"].__aexit__(None, None, None)
        if app.get("local_server"):
            stop_grpc_server(app["local_server"])


async def index(request):
    return aiohttp.web.Response(
        content_type="text/html",
        body="""<html>
<head><title>Software Heritage graph server</title></head>
<body>
<p>You have reached the <a href="https://www.softwareheritage.org/">
Software Heritage</a> graph API server.</p>

<p>See its
<a href="https://docs.softwareheritage.org/devel/swh-graph/api.html">API
documentation</a> for more information.</p>
</body>
</html>""",
    )


class GraphView(aiohttp.web.View):
    """Base class for views working on the graph, with utility functions"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rpc_client: TraversalServiceStub = self.request.app["rpc_client"]

    def get_direction(self):
        """Validate HTTP query parameter `direction`"""
        s = self.request.query.get("direction", "forward")
        if s not in ("forward", "backward"):
            raise aiohttp.web.HTTPBadRequest(text=f"invalid direction: {s}")
        return s.upper()

    def get_edges(self):
        """Validate HTTP query parameter `edges`, i.e., edge restrictions"""
        s = self.request.query.get("edges", "*")
        if any(
            [
                node_type != "*" and node_type not in EXTENDED_SWHID_TYPES
                for edge in s.split(":")
                for node_type in edge.split(",", maxsplit=1)
            ]
        ):
            raise aiohttp.web.HTTPBadRequest(text=f"invalid edge restriction: {s}")
        return s

    def get_return_types(self):
        """Validate HTTP query parameter 'return types', i.e,
        a set of types which we will filter the query results with"""
        s = self.request.query.get("return_types", "*")
        if any(
            node_type != "*" and node_type not in EXTENDED_SWHID_TYPES
            for node_type in s.split(",")
        ):
            raise aiohttp.web.HTTPBadRequest(
                text=f"invalid type for filtering res: {s}"
            )
        # if the user puts a star,
        # then we filter nothing, we don't need the other information
        if "*" in s:
            return "*"
        else:
            return s

    def get_max_matching_nodes(self):
        """Validate HTTP query parameter `max_matching_nodes`, i.e., number of results"""
        s = self.request.query.get("max_matching_nodes", "0")
        try:
            return int(s)
        except ValueError:
            raise aiohttp.web.HTTPBadRequest(
                text=f"invalid max_matching_nodes value: {s}"
            )

    def get_max_edges(self):
        """Validate HTTP query parameter 'max_edges', i.e.,
        the limit of the number of edges that can be visited"""
        s = self.request.query.get("max_edges", "0")
        try:
            return int(s)
        except ValueError:
            raise aiohttp.web.HTTPBadRequest(text=f"invalid max_edges value: {s}")

    async def check_swhid(self, swhid):
        """Validate that the given SWHID exists in the graph"""
        try:
            await self.rpc_client.GetNode(
                GetNodeRequest(swhid=swhid, mask=FieldMask(paths=["swhid"]))
            )
        except grpc.aio.AioRpcError as e:
            if e.code() in (
                grpc.StatusCode.INVALID_ARGUMENT,
                grpc.StatusCode.NOT_FOUND,  # Not used by Java backend, always "invalid"
            ):
                raise aiohttp.web.HTTPBadRequest(text=str(e.details()))


class StreamingGraphView(GraphView):
    """Base class for views streaming their response line by line."""

    content_type = "text/plain"

    @asynccontextmanager
    async def response_streamer(self, *args, **kwargs):
        """Context manager to prepare then close a StreamResponse"""
        response = aiohttp.web.StreamResponse(*args, **kwargs)
        response.content_type = self.content_type
        await response.prepare(self.request)
        yield response
        await response.write_eof()

    async def get(self):
        await self.prepare_response()
        async with self.response_streamer() as self.response_stream:
            self._buf = []
            try:
                await self.stream_response()
            finally:
                await self._flush_buffer()
            return self.response_stream

    async def prepare_response(self):
        """This can be overridden with some setup to be run before the response
        actually starts streaming.
        """
        pass

    async def stream_response(self):
        """Override this to perform the response streaming. Implementations of
        this should await self.stream_line(line) to write each line.
        """
        raise NotImplementedError

    async def stream_line(self, line):
        """Write a line in the response stream."""
        self._buf.append(line)
        if len(self._buf) > 100:
            await self._flush_buffer()

    async def _flush_buffer(self):
        await self.response_stream.write("\n".join(self._buf).encode() + b"\n")
        self._buf = []


class StatsView(GraphView):
    """View showing some statistics on the graph"""

    async def get(self):
        res = await self.rpc_client.Stats(StatsRequest())
        stats = json_format.MessageToDict(
            res,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
        )
        # Int64 fields are serialized as strings by default.
        for descriptor in res.DESCRIPTOR.fields:
            if descriptor.type == descriptor.TYPE_INT64:
                try:
                    stats[descriptor.name] = int(stats[descriptor.name])
                except KeyError:
                    pass
        json_body = json.dumps(stats, indent=4, sort_keys=True)
        return aiohttp.web.Response(body=json_body, content_type="application/json")


class SimpleTraversalView(StreamingGraphView):
    """Base class for views of simple traversals"""

    async def prepare_response(self):
        src = self.request.match_info["src"]
        self.traversal_request = TraversalRequest(
            src=[src],
            edges=self.get_edges(),
            direction=self.get_direction(),
            return_nodes=NodeFilter(types=self.get_return_types()),
            mask=FieldMask(paths=["swhid"]),
            max_matching_nodes=self.get_max_matching_nodes(),
        )
        if self.get_max_edges():
            self.traversal_request.max_edges = self.get_max_edges()
        await self.check_swhid(src)
        self.configure_request()
        self.nodes_stream = self.rpc_client.Traverse(
            self.traversal_request,
        )

        # Force gRPC to query the server and fetch the first nodes; so errors
        # are raised early, so we can return HTTP 503 before HTTP 200
        await self.nodes_stream.wait_for_connection()

    def configure_request(self):
        pass

    async def stream_response(self):
        async for node in self.nodes_stream:
            await self.stream_line(node.swhid)


class LeavesView(SimpleTraversalView):
    def configure_request(self):
        self.traversal_request.return_nodes.max_traversal_successors = 0


class NeighborsView(SimpleTraversalView):
    def configure_request(self):
        self.traversal_request.min_depth = 1
        self.traversal_request.max_depth = 1


class VisitNodesView(SimpleTraversalView):
    pass


class VisitEdgesView(SimpleTraversalView):
    def configure_request(self):
        self.traversal_request.mask.paths.extend(["successor.swhid"])
        # self.traversal_request.return_fields.successor = True

    async def stream_response(self):
        async for node in self.nodes_stream:
            for succ in node.successor:
                await self.stream_line(node.swhid + " " + succ.swhid)


class CountView(GraphView):
    """Base class for counting views."""

    count_type: Optional[str] = None

    async def get(self):
        src = self.request.match_info["src"]
        self.traversal_request = TraversalRequest(
            src=[src],
            edges=self.get_edges(),
            direction=self.get_direction(),
            return_nodes=NodeFilter(types=self.get_return_types()),
            mask=FieldMask(paths=["swhid"]),
            max_matching_nodes=self.get_max_matching_nodes(),
        )
        if self.get_max_edges():
            self.traversal_request.max_edges = self.get_max_edges()
        self.configure_request()
        res = await self.rpc_client.CountNodes(self.traversal_request)
        return aiohttp.web.Response(
            body=str(res.count), content_type="application/json"
        )

    def configure_request(self):
        pass


class CountNeighborsView(CountView):
    def configure_request(self):
        self.traversal_request.min_depth = 1
        self.traversal_request.max_depth = 1


class CountLeavesView(CountView):
    def configure_request(self):
        self.traversal_request.return_nodes.max_traversal_successors = 0


class CountVisitNodesView(CountView):
    pass


def make_app(config=None):
    """Create an aiohttp server for the HTTP RPC frontend to the swh-graph API.

    It may either connect to an existing grpc server (cls="remote") or spawn a
    local grpc server (cls="local").

    ``config`` is expected to be a dict like::

      graph:
        cls: "local"
        grpc_server:
          port: 50091
        http_rpc_server:
          debug: true

    or::

      graph:
        cls: "remote"
        url: "localhost:50091"
        http_rpc_server:
          debug: true

    See:

    - :mod:`swh.graph.grpc_server` for more details of the content of the
      grpc_server section,

    - :class:`~.GraphServerApp` class for more details of the content of the
      http_rpc_server section.

    """
    if config is None:
        config = {}
    if "graph" not in config:
        logger.info(
            "Missing 'graph' configuration; default to a locally spawn"
            "grpc server listening on 0.0.0.0:50091"
        )
        cfg = {"cls": "local", "grpc_server": {"port": 50091}}
    else:
        cfg = config["graph"].copy()
    cls = cfg.pop("cls")
    grpc_cfg = cfg.pop("grpc_server", {})
    app = GraphServerApp(**cfg.get("http_rpc_server", {}))
    if cls == "remote":
        if "url" not in cfg:
            raise KeyError("Missing 'url' configuration entry in the [graph] section")
        rpc_url = cfg["url"]
    elif cls in ("local", "local_rust"):
        app["local_server"], port = spawn_rust_grpc_server(**grpc_cfg)
        rpc_url = f"localhost:{port}"
    else:
        raise ValueError(f"Unknown swh.graph class cls={cls}")
    app.add_routes(
        [
            aiohttp.web.get("/", index),
            aiohttp.web.get("/graph", index),
            aiohttp.web.view("/graph/stats", StatsView),
            aiohttp.web.view("/graph/leaves/{src}", LeavesView),
            aiohttp.web.view("/graph/neighbors/{src}", NeighborsView),
            aiohttp.web.view("/graph/visit/nodes/{src}", VisitNodesView),
            aiohttp.web.view("/graph/visit/edges/{src}", VisitEdgesView),
            aiohttp.web.view("/graph/neighbors/count/{src}", CountNeighborsView),
            aiohttp.web.view("/graph/leaves/count/{src}", CountLeavesView),
            aiohttp.web.view("/graph/visit/nodes/count/{src}", CountVisitNodesView),
        ]
    )

    app["rpc_url"] = rpc_url
    return app


def make_app_from_configfile():
    """Load configuration and then build application to run"""
    config_file = os.environ.get("SWH_CONFIG_FILENAME")
    config = config_read(config_file)
    return make_app(config=config)



# Copyright (c) 2022-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib

import pytest

from swh.graph.grpc.swhgraph_pb2 import GraphDirection, NodeFilter, TraversalRequest

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)
TEST_ORIGIN_ID2 = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph2").hexdigest()
)


def get_leaves(
    graph_grpc_stub, src, direction=None, edges=None, max_matching_nodes=None
):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=src,
            direction=direction,
            edges=edges,
            max_matching_nodes=max_matching_nodes,
            return_nodes=NodeFilter(
                max_traversal_successors=0,
            ),
        )
    )
    return [node.swhid for node in request]


def _check_forward_from_snp(limit, actual_leaves):
    expected_leaves = [
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000007",
    ]

    if limit == 0:
        assert set(expected_leaves) == set(actual_leaves)
    else:
        assert set(actual_leaves) <= set(expected_leaves)
        assert len(actual_leaves) == min(limit, 4)


def test_forward_from_snp(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:snp:0000000000000000000000000000000000000020"],
        GraphDirection.FORWARD,
    )
    _check_forward_from_snp(0, actual_leaves)


@pytest.mark.parametrize("limit", [0, 1, 2, 4, 51 << 31 - 1])
def test_forward_from_snp_with_limit(graph_grpc_stub, limit):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:snp:0000000000000000000000000000000000000020"],
        max_matching_nodes=limit,
    )
    _check_forward_from_snp(limit, actual_leaves)


def test_forward_from_rel(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub, ["swh:1:rel:0000000000000000000000000000000000000019"]
    )
    expected_leaves = [
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000007",
        "swh:1:cnt:0000000000000000000000000000000000000011",
    ]
    assert set(expected_leaves) == set(actual_leaves)


def test_backward_from_leaf(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:cnt:0000000000000000000000000000000000000015"],
        direction=GraphDirection.BACKWARD,
    )
    expected_leaves = [
        "swh:1:rel:0000000000000000000000000000000000000019",
        TEST_ORIGIN_ID2,
    ]
    assert set(expected_leaves) == set(actual_leaves)

    actual_leaves2 = get_leaves(
        graph_grpc_stub,
        ["swh:1:cnt:0000000000000000000000000000000000000004"],
        direction=GraphDirection.BACKWARD,
    )
    expected_leaves2 = [
        TEST_ORIGIN_ID,
        TEST_ORIGIN_ID2,
        "swh:1:rel:0000000000000000000000000000000000000019",
    ]
    assert set(expected_leaves2) == set(actual_leaves2)


def test_forward_rev_to_rev_only(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:rev:0000000000000000000000000000000000000018"],
        edges="rev:rev",
    )
    expected_leaves = ["swh:1:rev:0000000000000000000000000000000000000003"]
    assert set(expected_leaves) == set(actual_leaves)


def test_forward_dir_to_all(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:dir:0000000000000000000000000000000000000008"],
        edges="dir:*",
    )
    expected_leaves = [
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000007",
    ]
    assert set(expected_leaves) == set(actual_leaves)


def test_backward_cnt_to_dir_dir_to_dir(graph_grpc_stub):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:cnt:0000000000000000000000000000000000000005"],
        edges="cnt:dir,dir:dir",
        direction=GraphDirection.BACKWARD,
    )
    expected_leaves = ["swh:1:dir:0000000000000000000000000000000000000012"]
    assert set(expected_leaves) == set(actual_leaves)


@pytest.mark.parametrize("limit", [0, 1, 2, 1 << 31 - 1])
def test_backward_cnt_to_dir_dir_to_dir_with_limit(graph_grpc_stub, limit):
    actual_leaves = get_leaves(
        graph_grpc_stub,
        ["swh:1:cnt:0000000000000000000000000000000000000005"],
        edges="cnt:dir,dir:dir",
        direction=GraphDirection.BACKWARD,
        max_matching_nodes=limit,
    )
    expected_leaves = ["swh:1:dir:0000000000000000000000000000000000000012"]
    assert set(expected_leaves) == set(actual_leaves)
